# -*- coding: utf-8 -*-
from __future__ import print_function
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import unittest

import qq0196 as qq
import qu0196 as qu
import uq0196 as uq
import uu0196 as uu
import mix
from pyxb.utils.domutils import BindingDOMSupport
from pyxb.utils import six

BindingDOMSupport.DeclareNamespace(qq.Namespace, 'qq')
BindingDOMSupport.DeclareNamespace(qu.Namespace, 'qu')
BindingDOMSupport.DeclareNamespace(uq.Namespace, 'uq')
BindingDOMSupport.DeclareNamespace(uu.Namespace, 'uu')
BindingDOMSupport.DeclareNamespace(mix.Namespace, 'mix')

qq_bds = BindingDOMSupport(default_namespace=qq.Namespace)


elt_kw = {
    'te' : 'te',
    'teq' : 'teq',
    'teu' : 'teu',
    'e' : 'e',
    'eq' : 'eq',
    'eu' : 'eu',
    'a' : 'a',
    'aq' : 'aq',
    'au' : 'au',
    'ta' : 'ta',
    'taq' : 'taq',
    'tau' : 'tau' }

qq_i = qq.elt(**elt_kw)
qu_i = qu.elt(**elt_kw)
uq_i = uq.elt(**elt_kw)
uu_i = uu.elt(**elt_kw)

i = mix.elt(qq_i, qu_i, uq_i, uu_i)
try:
    print(i.toDOM().toprettyxml())
except pyxb.ValidationError as e:
    print(e.details())
    raise

i = mix.uue(a='a')
print(i.toxml('utf-8'))

class TestTrac0196 (unittest.TestCase):

    module_map = { qq : ( qq.Namespace, qq.Namespace ),
                   qu : ( qu.Namespace, None ),
                   uq : ( None, uq.Namespace ),
                   uu : ( None, None ) }
    global_a = ( 'a', 'aq', 'au' )
    global_e = ('e', 'eq', 'eu' )
    local_a = ( 'ta', 'taq', 'tau' )
    local_e = ('te', 'teq', 'teu' )

    def testQualified (self):
        # Top-level declarations are qualified regardless of presence/absence of form attribute.
        # Internal declarations follow form attribute or schema default
        for (m, ( efd, afd )) in six.iteritems(self.module_map):
            for (n, d) in six.iteritems(m.t._AttributeMap):
                if n.localName() in ('a', 'au', 'aq'):
                    self.assertEqual(n.namespace(), m.Namespace)
                elif 'taq' == n.localName():
                    self.assertEqual(n.namespace(), m.Namespace)
                elif 'tau' == n.localName():
                    self.assertEqual(n.namespace(), None)
                elif 'ta' == n.localName():
                    self.assertEqual(n.namespace(), afd)
                else:
                    self.assertFalse()
            for (n, d) in six.iteritems(m.t._ElementMap):
                if n.localName() in ('e', 'eu', 'eq'):
                    self.assertEqual(n.namespace(), m.Namespace)
                elif 'teq' == n.localName():
                    self.assertEqual(n.namespace(), m.Namespace)
                elif 'teu' == n.localName():
                    self.assertEqual(n.namespace(), None)
                elif 'te' == n.localName():
                    self.assertEqual(n.namespace(), efd)
                else:
                    self.assertFalse()

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.utils.domutils
from xml.dom import Node

import os.path
xsd='''<?xml version="1.0" encoding="UTF-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
  <xs:element name="a" type="xs:string"/>
  <xs:complexType name="mType">
   <xs:sequence>
     <xs:element ref="a" minOccurs="0" maxOccurs="unbounded"/>
   </xs:sequence>
  </xs:complexType>
  <xs:element name="m" type="mType"/>
</xs:schema>
'''

code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
#open('code.py', 'w').write(code)
#print code

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

class TestTrac0201 (unittest.TestCase):
    def testBasicList (self):
        lst = []
        lst.insert(0, 'one')
        lst.insert(0, 'zero')
        self.assertEqual(lst, ['zero', 'one'])

    def testElementList (self):
        i = m()
        i.a.insert(0, 'one')
        i.a.insert(0, 'zero')
        self.assertEqual(i.a, ['zero', 'one'])

if __name__ == '__main__':
    unittest.main()



# Bring everything in
from raw.ipo import *
from pyxb.utils import six

# Provide a way to reference the raw bindings that we're going to override
import raw.ipo as raw_ipo

@six.python_2_unicode_compatible
class USAddress (raw_ipo.USAddress):
    def __str__ (self):
        return six.u('''%s
%s
%s %s, %s''') % (self.name(), self.street(), self.city(), self.state(), self.zip())
    pass
raw_ipo.USAddress._SetSupersedingClass(USAddress)



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.binding.datatypes as xs
import pyxb.binding.basis
import pyxb.utils.domutils

import os.path
xsd='''<?xml version="1.0" encoding="UTF-8"?>
<xs:schema targetNamespace="whatever"
  xmlns:whatever="whatever"
  xmlns:xs="http://www.w3.org/2001/XMLSchema">

 <xs:attribute name="lang">
  <xs:simpleType>
   <xs:union memberTypes="xs:language">
    <xs:simpleType>
     <xs:restriction base="xs:string">
      <xs:enumeration value=""/>
     </xs:restriction>
    </xs:simpleType>
   </xs:union>
  </xs:simpleType>
 </xs:attribute>
</xs:schema>
'''

code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
#open('code.py', 'w').write(code)
#print code

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

class TestBug_200907251353 (unittest.TestCase):
    def testBasic (self):
        anon = [ _s for _s in globals() if _s.startswith('STD_ANON') ]
        self.assertEqual(2, len(anon))

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import unittest
import pyxb

class TestTrac0199 (unittest.TestCase):

    def setUp (self):
        self.__pyxb_version = pyxb.__version__
        pyxb.__version__ = 'NOT ' + self.__pyxb_version

    def tearDown (self):
        pyxb.__version__ = self.__pyxb_version

    def testImport (self):
        try:
            import X
            self.assertFalse('exception not raised')
        except pyxb.PyXBVersionError as e:
            self.assertEqual(e.args[0], self.__pyxb_version)

if __name__ == '__main__':
    unittest.main()



from __future__ import print_function
import address

addr = address.USAddress()
addr.name = 'Robert Smith'
addr.street = '8 Oak Avenue'
addr.city = 'Anytown'
addr.state = 'NY'
addr.zip = 12341

print(addr.toxml("utf-8"))



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb
import unittest
import datetime
import pyxb.binding.datatypes as xsd

class Test_duration (unittest.TestCase):
    def testBasic (self):
        v = xsd.duration('P1347Y')
        self.assertEqual(0, v.days)
        self.assertEqual(0, v.seconds)
        self.assertEqual(0, v.microseconds)
        self.assertEqual(1347, v.durationData()['years'])
        self.assertEqual('P1347Y', v.xsdLiteral())

        v = xsd.duration('P1347M')
        self.assertEqual(0, v.days)
        self.assertEqual(0, v.seconds)
        self.assertEqual(0, v.microseconds)
        self.assertEqual(1347, v.durationData()['months'])
        self.assertEqual('P1347M', v.xsdLiteral())

        v = xsd.duration('P1Y2MT2H')
        self.assertEqual(0, v.days)
        self.assertEqual(2 * 60 * 60, v.seconds)
        self.assertEqual(0, v.microseconds)
        self.assertEqual(1, v.durationData()['years'])
        self.assertEqual(2, v.durationData()['months'])
        self.assertEqual('P1Y2MT2H', v.xsdLiteral())

        v = xsd.duration('P0Y1347M')
        self.assertEqual(0, v.days)
        self.assertEqual(0, v.seconds)
        self.assertEqual(0, v.microseconds)
        self.assertEqual(1347, v.durationData()['months'])
        self.assertEqual('P1347M', v.xsdLiteral())

        v = xsd.duration('P0Y1347M0D')
        self.assertEqual(0, v.days)
        self.assertEqual(0, v.seconds)
        self.assertEqual(0, v.microseconds)
        self.assertEqual(1347, v.durationData()['months'])
        self.assertFalse(v.negativeDuration())
        self.assertEqual('P1347M', v.xsdLiteral())

        self.assertRaises(pyxb.SimpleTypeValueError, xsd.duration, 'P-1347M')
        v = xsd.duration('-P1347M')
        self.assertEqual(0, v.days)
        self.assertEqual(0, v.seconds)
        self.assertEqual(0, v.microseconds)
        self.assertEqual(1347, v.durationData()['months'])
        self.assertTrue(v.negativeDuration())
        self.assertEqual('-P1347M', v.xsdLiteral())

        self.assertRaises(pyxb.SimpleTypeValueError, xsd.duration, 'P1Y2MT')

        v = xsd.duration('P3DT4H7M')

    def testFromTimeDelta (self):
        v = xsd.duration(datetime.timedelta(seconds=3))
        self.assertEqual('PT3S', v.xsdLiteral())
        v = xsd.duration(datetime.timedelta(minutes=3))
        self.assertEqual('PT3M', v.xsdLiteral())
        v = xsd.duration(datetime.timedelta(hours=3))
        self.assertEqual('PT3H', v.xsdLiteral())
        v = xsd.duration(datetime.timedelta(hours=3, seconds=61))
        self.assertEqual('PT3H1M1S', v.xsdLiteral())
        v = xsd.duration(datetime.timedelta(hours=27, seconds=61))
        self.assertEqual('P1DT3H1M1S', v.xsdLiteral())
        v = xsd.duration(datetime.timedelta(seconds=61+27*60*60))
        self.assertEqual('P1DT3H1M1S', v.xsdLiteral())

    def testNegative (self):
        v = xsd.duration('-P3DT4H7M23.5S')
        # Time is 19H52M36.5S into day -4
        base_date = xsd.dateTime('2000-01-10T00:00:00')
        delta_date = xsd.dateTime(base_date + v)
        self.assertEqual(-4, v.days)
        self.assertEqual(36 + 60 * (52 + 60 * 19), v.seconds)
        self.assertEqual(500000, v.microseconds)
        self.assertEqual('2000-01-06T19:52:36.5', delta_date.xsdLiteral())
        td = datetime.timedelta(days=-1)
        v = xsd.duration(td)
        self.assertEqual(v, td)
        delta_date = xsd.dateTime(base_date + v)
        self.assertEqual('2000-01-09T00:00:00', delta_date.xsdLiteral())

    def testZero (self):
        xmlt = 'P0D';
        v = xsd.duration(xmlt)
        self.assertEqual(0, v.total_seconds())
        self.assertEqual(xmlt, v.xsdLiteral())

    def testAddition (self):
        date = xsd.dateTime(2002, 10, 27, 12, 14, 32)
        duration = xsd.duration('P3DT5H3M')
        self.assertEqual('2002-10-30T17:17:32', xsd.dateTime(date + duration).xsdLiteral())
        self.assertEqual('2002-10-24T07:11:32', xsd.dateTime(date - duration).xsdLiteral())

    def testCreation (self):
        base_date = xsd.dateTime('2000-01-10T00:00:00')
        delta_date = xsd.dateTime('2000-01-06T19:52:37.5')
        v = xsd.duration(base_date - delta_date)
        self.assertEqual(3, v.days)
        self.assertEqual(14842, v.seconds)
        self.assertEqual('P3DT4H7M22.5S', v.xsdLiteral())
        self.assertRaises(pyxb.SimpleTypeValueError, xsd.duration)
        self.assertRaises(pyxb.SimpleTypeValueError, xsd.duration, 4)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
from pyxb.exceptions_ import *
import unittest
import pyxb.binding.datatypes as xsd

class Test_ENTITIES (unittest.TestCase):
    def testBasicLists (self):
        v = xsd.ENTITIES([ "one", "two", "three" ])
        self.assertEqual(3, len(v))
        self.assertTrue(isinstance(v[0], xsd.ENTITY))
        self.assertEqual("one", v[0])

    def testStringLists (self):
        v = xsd.ENTITIES("one two three")
        self.assertEqual(3, len(v))
        self.assertEqual("one", v[0])
        self.assertTrue(isinstance(v[0], xsd.ENTITY))
        self.assertRaises(SimpleTypeValueError, xsd.ENTITIES, 'string with b@d id')

if __name__ == '__main__':
    unittest.main()



from __future__ import print_function
import dict
from pyxb.utils.six.moves.urllib.request import urlopen
import pyxb.utils.domutils as domutils
from xml.dom import minidom

# Get the list of dictionaries available from the service.
port_uri = 'http://services.aonaware.com/DictService/DictService.asmx'
uri = port_uri + '/DictionaryList'
dle_xml = urlopen(uri).read()
dle_dom = domutils.StringToDOM(dle_xml)
dle = dict.ArrayOfDictionary.createFromDOM(dle_dom)

op_path = '/DictionaryInfo'
for d in dle.Dictionary:
    # Create a REST-style query to retrieve the information about this dictionary.
    uri = '%s%s?dictId=%s' % (port_uri, op_path, d.Id)
    resp = urlopen(uri).read()
    # The response is a simple type derived from string, so we can
    # just extract and print it.
    di_resp = dict.CreateFromDOM(domutils.StringToDOM(resp))
    # Do the "encode" garbage because one of these dictionaries has a
    # non-ASCII character
    print("%s (%s)\n%s\n" % (d.Name.encode('utf-8'), d.Id.encode('utf-8'), di_resp.encode('utf-8')))



# Copyright 2009-2013, Peter A. Bigot
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain a
# copy of the License at:
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Utility to generate the code point sets defined by the Unicode standard.

You'll need these files, corresponding to the U{Unicode Database for XML
Schema Part 2: Datatypes<http://www.w3.org/TR/xmlschema-2/#UnicodeDB>},
which is:

 - U{http://www.unicode.org/Public/3.1-Update/UnicodeData-3.1.0.txt}
 - U{http://www.unicode.org/Public/3.1-Update/Blocks-4.txt}

Invoke this script, redirecting the output to
C{pyxb/utils/unicode_data.py}.

"""

from __future__ import print_function
import textwrap
import re
from pyxb.utils import six
from pyxb.utils.six.moves import xrange

def countCodepoints (codepoints):
    count = 0
    for v in codepoints:
        if isinstance(v, tuple):
            count = count + v[1] - v[0] + 1
        else:
            count = count + 1
    return count

def condenseCodepoints (codepoints):
    ranges = []
    codepoints = list(codepoints)
    codepoints.sort()
    range_min = None
    for ri in xrange(len(codepoints)):
        codepoint = codepoints[ri]
        if not isinstance(codepoint, tuple):
            if range_min is None:
                range_last = range_min = codepoint
                range_next = range_last + 1
                continue
            if codepoint == range_next:
                range_last = codepoint
                range_next += 1
                continue
        if range_min is not None:
            ranges.append( (range_min, range_last) )
        if isinstance(codepoint, tuple):
            range_min = None
            ranges.append(codepoint)
        else:
            range_last = range_min = codepoints[ri]
            range_next = range_last + 1
    if range_min is not None:
        ranges.append( (range_min, range_last) )
    return ranges

def rangesToPython (ranges, indent=11, width=67):
    ranges.sort()
    text = ', '.join( [ '(0x%06x, 0x%06x)' % _r for _r in ranges ] )
    text += ','
    wrapped = textwrap.wrap(text, 67)
    return ("\n%s" % (' ' * indent,)).join(wrapped)

def emitCategoryMap (data_file):
    category_map = {}
    unicode_data = open(data_file)
    range_first = None
    last_codepoint = -1
    while True:
        line = unicode_data.readline()
        fields = line.split(';')
        if 1 >= len(fields):
            break
        codepoint = int(fields[0], 16)
        char_name = fields[1]
        category = fields[2]

        # If code points are are not listed in the file, they are in the Cn category.
        if range_first is None and last_codepoint + 1 != codepoint:
            category_map.setdefault('Cn', []).append((last_codepoint + 1, codepoint))
            category_map.setdefault('C', []).append((last_codepoint + 1, codepoint))
        last_codepoint = codepoint

        if char_name.endswith(', First>'):
            assert range_first is None
            range_first = codepoint
            continue
        if range_first is not None:
            assert char_name.endswith(', Last>')
            codepoint = ( range_first, codepoint )
            range_first = None
        category_map.setdefault(category, []).append(codepoint)
        category_map.setdefault(category[0], []).append(codepoint)

    # Code points at the end of the Unicode range that are are not listed in
    # the file are in the Cn category.
    category_map.setdefault('Cn', []).append((last_codepoint + 1, 0x10FFFF))
    category_map.setdefault('C', []).append((last_codepoint + 1, 0x10FFFF))

    for k, v in six.iteritems(category_map):
        category_map[k] = condenseCodepoints(v)

    print('# Unicode general category properties: %d properties' % (len(category_map),))
    print('PropertyMap = {')
    for (k, v) in sorted(six.iteritems(category_map)):
        print('  # %s: %d codepoint groups (%d codepoints)' % (k, len(v), countCodepoints(v)))
        print("  %-4s : CodePointSet([" % ("'%s'" % k,))
        print("           %s" % (rangesToPython(v, indent=11, width=67),))
        print("         ]),")
    print('  }')

def emitBlockMap (data_file):
    block_map = { }
    block_re = re.compile('(?P<min>[0-9A-F]+)(?P<spans>\.\.|; )(?P<max>[0-9A-F]+);\s(?P<block>.*)$')
    block_data = open(data_file)
    while True:
        line = block_data.readline()
        if 0 == len(line):
            break
        mo = block_re.match(line)
        if mo is None:
            continue
        rmin = int(mo.group('min'), 16)
        rmax = int(mo.group('max'), 16)
        block = mo.group('block').replace(' ', '')
        block_map.setdefault(block, []).append( (rmin, rmax) )

    print('# Unicode code blocks: %d blocks' % (len(block_map),))
    print('BlockMap = {')
    for k in sorted(six.iterkeys(block_map)):
        v = block_map.get(k)
        print('  %s : CodePointSet(' % (repr(k),))
        print('     %s' % (rangesToPython(v, indent=6, width=67),))
        print('  ),')
    print('  }')

print('''# -*- coding: utf-8 -*-
# Unicode property and category maps.

from pyxb.utils.unicode import CodePointSet
''')

emitBlockMap('Blocks-4.txt')
emitCategoryMap('UnicodeData-3.1.0.txt')



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.binding.datatypes as xs
import pyxb.binding.basis
import pyxb.utils.domutils

import os.path
xsd='''
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
    <xs:element name="wbin">
        <xs:complexType>
            <xs:sequence>
                <xs:element name="bin" type="xs:hexBinary"/>
            </xs:sequence>
        </xs:complexType>
    </xs:element>
    <xs:element name="bin" type="xs:hexBinary"/>
    <xs:element name="u64" type="xs:base64Binary"/>
</xs:schema>
'''

#open('schema.xsd', 'w').write(xsd)
code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
#open('code.py', 'w').write(code)
#print code

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

class TestIssue0051 (unittest.TestCase):
    def testWrapped (self):
        i = CreateFromDocument('<wbin><bin>aa</bin></wbin>')
        self.assertEqual(1, len(i.bin))
        i = CreateFromDocument('<wbin><bin> </bin></wbin>')
        self.assertEqual(0, len(i.bin))
        i = CreateFromDocument('<wbin><bin/></wbin>')
        self.assertEqual(0, len(i.bin))

    def testBare (self):
        i = CreateFromDocument('<bin>aa</bin>')
        self.assertEqual(1, len(i))
        i = CreateFromDocument('<bin> </bin>')
        self.assertEqual(0, len(i))
        i = CreateFromDocument('<bin/>')
        self.assertEqual(0, len(i))
        i = CreateFromDocument('<u64>Zg==</u64>')
        self.assertEqual(1, len(i))
        i = CreateFromDocument('<u64> </u64>')
        self.assertEqual(0, len(i))
        i = CreateFromDocument('<u64/>')
        self.assertEqual(0, len(i))

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright 2009-2013, Peter A. Bigot
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain a
# copy of the License at:
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Utility functions and classes."""

import re
import os
import errno
import pyxb
from pyxb.utils.six.moves.urllib import parse as urlparse
import time
import datetime
import logging
from pyxb.utils import six

_log = logging.getLogger(__name__)

class Object (object):
    """A dummy class used to hold arbitrary attributes.

    Essentially this gives us a map without having to worry about
    converting names to text to use as keys.
    """
    pass

def BackfillComparisons (cls):
    """Class decorator that fills in missing ordering methods.

    Concept derived from Python 2.7.5 functools.total_ordering,
    but this version requires that __eq__ and __lt__ be provided,
    and unconditionally overrides __ne__, __gt__, __le__, and __ge__
    with the derived versions.

    This is still necessary in Python 3 because in Python 3 the
    comparison x >= y is done by the __ge__ inherited from object,
    which does not handle the case where x and y are not the same type
    even if the underlying y < x would convert x to be compatible. """

    def applyconvert (cls, derived):
        for (opn, opx) in derived:
            opx.__name__ = opn
            opx.__doc__ = getattr(int, opn).__doc__
            setattr(cls, opn, opx)

    applyconvert(cls, (
            ('__gt__', lambda self, other: not (self.__lt__(other) or self.__eq__(other))),
            ('__le__', lambda self, other: self.__lt__(other) or self.__eq__(other)),
            ('__ge__', lambda self, other: not self.__lt__(other))
            ))
    applyconvert(cls, (
            ('__ne__', lambda self, other: not self.__eq__(other)),
            ))
    return cls

def IteratedCompareMixed (lhs, rhs):
    """Tuple comparison that permits C{None} as lower than any value,
    and defines other cross-type comparison.

    @return: -1 if lhs < rhs, 0 if lhs == rhs, 1 if lhs > rhs."""
    li = iter(lhs)
    ri = iter(rhs)
    while True:
        try:
            (lv, rv) = (next(li), next(ri))
            if lv is None:
                if rv is None:
                    continue
                return -1
            if rv is None:
                return 1
            if lv == rv:
                continue
            if lv < rv:
                return -1
            return 1
        except StopIteration:
            nl = len(lhs)
            nr = len(rhs)
            if nl < nr:
                return -1
            if nl == nr:
                return 0
            return 1

def QuotedEscaped (s):
    """Convert a string into a literal value that can be used in Python source.

    This just calls C{repr}.  No point in getting all complex when the language
    already gives us what we need.

    @rtype: C{str}
    """
    return repr(s)

def _DefaultXMLIdentifierToPython (identifier):
    """Default implementation for _XMLIdentifierToPython

    For historical reasons, this converts the identifier from a str to
    unicode in the system default encoding.  This should have no
    practical effect.

    @param identifier : some XML identifier

    @return: C{unicode(identifier)}
    """

    return six.text_type(identifier)

def _SetXMLIdentifierToPython (xml_identifier_to_python):
    """Configure a callable L{MakeIdentifier} uses to pre-process an XM Lidentifier.

    In Python3, identifiers can be full Unicode tokens, but in Python2,
    all identifiers must be ASCII characters.  L{MakeIdentifier} enforces
    this by removing all characters that are not valid within an
    identifier.

    In some cases, an application generating bindings may be able to
    transliterate Unicode code points that are not valid Python identifier
    characters into something else.  This callable can be assigned to
    perform that translation before the invalid characters are
    stripped.

    It is not the responsibility of this callable to do anything other
    than replace whatever characters it wishes to.  All
    transformations performed by L{MakeIdentifier} will still be
    applied, to ensure the output is in fact a legal identifier.

    @param xml_identifier_to_python : A callable that takes a string
    and returns a Unicode, possibly with non-identifier characters
    replaced by other characters.  Pass C{None} to reset to the
    default implementation, which is L{_DefaultXMLIdentifierToPython}.

    @rtype: C{unicode}
    """
    global _XMLIdentifierToPython
    if xml_identifier_to_python is None:
        xml_identifier_to_python = _DefaultXMLIdentifierToPython
    _XMLIdentifierToPython = xml_identifier_to_python

_XMLIdentifierToPython = _DefaultXMLIdentifierToPython

_UnderscoreSubstitute_re = re.compile(r'[- .]')
_NonIdentifier_re = re.compile(r'[^a-zA-Z0-9_]')
_PrefixUnderscore_re = re.compile(r'^_+')
_PrefixDigit_re = re.compile(r'^\d+')
_CamelCase_re = re.compile(r'_\w')

def MakeIdentifier (s, camel_case=False):
    """Convert a string into something suitable to be a Python identifier.

    The string is processed by L{_XMLIdentifierToPython}.  Following
    this, dashes, spaces, and periods are replaced by underscores, and
    characters not permitted in Python identifiers are stripped.
    Furthermore, any leading underscores are removed.  If the result
    begins with a digit, the character 'n' is prepended.  If the
    result is the empty string, the string 'emptyString' is
    substituted.

    No check is made for L{conflicts with keywords <DeconflictKeyword>}.

    @keyword camel_case : If C{True}, any underscore in the result
    string that is immediately followed by an alphanumeric is replaced
    by the capitalized version of that alphanumeric.  Thus,
    'one_or_two' becomes 'oneOrTwo'.  If C{False} (default), has no
    effect.

    @rtype: C{str}
    """
    s = _XMLIdentifierToPython(s)
    s = _PrefixUnderscore_re.sub('', _NonIdentifier_re.sub('', _UnderscoreSubstitute_re.sub('_', s)))
    if camel_case:
        s = _CamelCase_re.sub(lambda _m: _m.group(0)[1].upper(), s)
    if _PrefixDigit_re.match(s):
        s = 'n' + s
    if 0 == len(s):
        s = 'emptyString'
    return s

def MakeModuleElement (s):
    """Convert a string into something that can be a valid element in a
    Python module path.

    Module path elements are similar to identifiers, but may begin
    with numbers and should not have leading underscores removed.
    """
    return _UnderscoreSubstitute_re.sub('_', _XMLIdentifierToPython(s))

_PythonKeywords = frozenset( (
        "and", "as", "assert", "break", "class", "continue", "def", "del",
        "elif", "else", "except", "exec", "finally", "for", "from", "global",
        "if", "import", "in", "is", "lambda", "not", "or", "pass", "print",
        "raise", "return", "try", "while", "with", "yield"
        ) )
"""Python keywords.  Note that types like int and float are not
keywords.

@see: U{http://docs.python.org/reference/lexical_analysis.html#keywords}."""

_PythonBuiltInConstants = frozenset( (
        "False", "True", "None", "NotImplemented", "Ellipsis", "__debug__",
        # "set" is neither a keyword nor a constant, but if some fool
        # like {http://www.w3.org/2001/SMIL20/}set gets defined there's
        # no way to access the builtin constructor.
        "set"
        ) )
"""Other symbols that aren't keywords but that can't be used.

@see: U{http://docs.python.org/library/constants.html}."""

_Keywords = frozenset(_PythonKeywords.union(_PythonBuiltInConstants))
"""The keywords reserved for Python, derived from L{_PythonKeywords}
and L{_PythonBuiltInConstants}."""

def DeconflictKeyword (s, aux_keywords=frozenset()):
    """If the provided string C{s} matches a Python language keyword,
    append an underscore to distinguish them.

    See also L{MakeUnique}.

    @param s: string to be deconflicted

    @keyword aux_keywords: optional iterable of additional strings
    that should be treated as keywords.

    @rtype: C{str}

    """
    if (s in _Keywords) or (s in aux_keywords):
        return '%s_' % (s,)
    return s

def MakeUnique (s, in_use):
    """Return an identifier based on C{s} that is not in the given set.

    The returned identifier is made unique by appending an underscore
    and, if necessary, a serial number.

    The order is : C{x}, C{x_}, C{x_2}, C{x_3}, ...

    @param in_use: The set of identifiers already in use in the
    relevant scope.  C{in_use} is updated to contain the returned
    identifier.

    @rtype: C{str}
    """
    if s in in_use:
        ctr = 2
        s = s.rstrip('_')
        candidate = '%s_' % (s,)
        while candidate in in_use:
            candidate = '%s_%d' % (s, ctr)
            ctr += 1
        s = candidate
    in_use.add(s)
    return s

def PrepareIdentifier (s, in_use, aux_keywords=frozenset(), private=False, protected=False):
    """Combine everything required to create a unique identifier.

    Leading and trailing underscores are stripped from all
    identifiers.

    @param in_use: the set of already used identifiers.  Upon return
    from this function, it is updated to include the returned
    identifier.

    @keyword aux_keywords: an optional set of additional symbols that
    are illegal in the given context; use this to prevent conflicts
    with known method names.

    @keyword private: if C{False} (default), all leading underscores
    are stripped, guaranteeing the identifier will not be private.  If
    C{True}, the returned identifier has two leading underscores,
    making it a private variable within a Python class.

    @keyword protected: as for C{private}, but uses only one
    underscore.

    @rtype: C{str}

    @note: Only module-level identifiers should be treated as
    protected.  The class-level L{_DeconflictSymbols_mixin}
    infrastructure does not include protected symbols.  All class and
    instance members beginning with a single underscore are reserved
    for the PyXB infrastructure."""
    s = DeconflictKeyword(MakeIdentifier(s).strip('_'), aux_keywords)
    if private:
        s = '__' + s
    elif protected:
        s = '_' + s
    return MakeUnique(s, in_use)

# @todo: descend from pyxb.cscRoot, if we import pyxb
class _DeconflictSymbols_mixin (object):
    """Mix-in used to deconflict public symbols in classes that may be
    inherited by generated binding classes.

    Some classes, like the L{pyxb.binding.basis.element} or
    L{pyxb.binding.basis.simpleTypeDefinition} classes in
    L{pyxb.binding.basis}, have public symbols associated with
    functions and variables.  It is possible that an XML schema might
    include tags and attribute names that match these symbols.  To
    avoid conflict, the reserved symbols marked in this class are
    added to the pre-defined identifier set.

    Subclasses should create a class-level variable
    C{_ReservedSymbols} that contains a set of strings denoting the
    symbols reserved in this class, combined with those from any
    superclasses that also have reserved symbols.  Code like the
    following is suggested::

       # For base classes (direct mix-in):
       _ReservedSymbols = set([ 'one', 'two' ])
       # For subclasses:
       _ReservedSymbols = SuperClass._ReservedSymbols.union(set([ 'three' ]))

    Only public symbols (those with no underscores) are currently
    supported.  (Private symbols can't be deconflicted that easily,
    and no protected symbols that derive from the XML are created by
    the binding generator.)
    """

    _ReservedSymbols = set()
    """There are no reserved symbols in the base class."""

# Regular expression detecting tabs, carriage returns, and line feeds
__TabCRLF_re = re.compile("[\t\n\r]")
# Regular expressoin detecting sequences of two or more spaces
__MultiSpace_re = re.compile(" +")

def NormalizeWhitespace (text, preserve=False, replace=False, collapse=False):
    """Normalize the given string.

    Exactly one of the C{preserve}, C{replace}, and C{collapse} keyword
    parameters must be assigned the value C{True} by the caller.

     - C{preserve}: the text is returned unchanged.

     - C{replace}: all tabs, newlines, and carriage returns are
     replaced with ASCII spaces.

     - C{collapse}: the C{replace} normalization is done, then
     sequences of two or more spaces are replaced by a single space.

    See the U{whiteSpace facet<http://www.w3.org/TR/xmlschema-2/#rf-whiteSpace>}.

    @rtype: C{str}
    """
    if preserve:
        return text
    text = __TabCRLF_re.sub(' ', text)
    if replace:
        return text
    if collapse:
        return __MultiSpace_re.sub(' ', text).strip()
    # pyxb not imported here; could be.
    raise Exception('NormalizeWhitespace: No normalization specified')

class Graph:
    """Represent a directed graph with arbitrary objects as nodes.

    This is used in the L{code
    generator<pyxb.binding.generate.Generator>} to determine order
    dependencies among components within a namespace, and schema that
    comprise various namespaces.  An edge from C{source} to C{target}
    indicates that some aspect of C{source} requires that some aspect
    of C{target} already be available.
    """

    def __init__ (self, root=None):
        self.__roots = None
        if root is not None:
            self.__roots = set([root])
        self.__edges = set()
        self.__edgeMap = { }
        self.__reverseMap = { }
        self.__nodes = set()

    __scc = None
    __sccMap = None
    __dfsOrder = None

    def addEdge (self, source, target):
        """Add a directed edge from the C{source} to the C{target}.

        The nodes are added to the graph if necessary.
        """
        self.__edges.add( (source, target) )
        self.__edgeMap.setdefault(source, set()).add(target)
        if source != target:
            self.__reverseMap.setdefault(target, set()).add(source)
        self.__nodes.add(source)
        self.__nodes.add(target)

    def addNode (self, node):
        """Add  the given node to the graph."""
        self.__nodes.add(node)

    __roots = None
    def roots (self, reset=False):
        """Return the set of nodes calculated to be roots (i.e., those that have no incoming edges).

        This caches the roots calculated in a previous invocation
        unless the C{reset} keyword is given the value C{True}.

        @note: Upon reset, any notes that had been manually added
        using L{addNode} will no longer be in the set.

        @keyword reset: If C{True}, any cached value is discarded and
        recomputed.  No effect if C{False} (defalut).

        @rtype: C{set}
        """
        if reset or (self.__roots is None):
            self.__roots = set()
            for n in self.__nodes:
                if not (n in self.__reverseMap):
                    self.__roots.add(n)
        return self.__roots
    def addRoot (self, root):
        """Add the provided node as a root node, even if it has incoming edges.

        The node need not be present in the graph (if necessary, it is added).

        Note that roots added in this way do not survive a reset using
        L{roots}.

        @return: C{self}
        """
        if self.__roots is None:
            self.__roots = set()
        self.__nodes.add(root)
        self.__roots.add(root)
        return self

    def edgeMap (self):
        """Return the edges in the graph.

        The edge data structure is a map from the source node to the
        set of nodes that can be reached in a single step from the
        source.
        """
        return self.__edgeMap
    __edgeMap = None

    def edges (self):
        """Return the edges in the graph.

        The edge data structure is a set of node pairs represented as C{( source, target )}.
        """
        return self.__edges

    def nodes (self):
        """Return the set of nodes in the graph.

        The node collection data structure is a set containing node
        objects, whatever they may be."""
        return self.__nodes

    def tarjan (self, reset=False):
        """Execute Tarjan's algorithm on the graph.

        U{Tarjan's
        algorithm<http://en.wikipedia.org/wiki/Tarjan%27s_strongly_connected_components_algorithm>}
        computes the U{strongly-connected
        components<http://en.wikipedia.org/wiki/Strongly_connected_component>}
        of the graph: i.e., the sets of nodes that form a minimal
        closed set under edge transition.  In essence, the loops.  We
        use this to detect groups of components that have a dependency
        cycle.

        @keyword reset: If C{True}, any cached component set is erased
        and recomputed.  If C{True}, an existing previous result is
        left unchanged."""

        if (self.__scc is not None) and (not reset):
            return
        self.__sccMap = { }
        self.__stack = []
        self.__sccOrder = []
        self.__scc = []
        self.__index = 0
        self.__tarjanIndex = { }
        self.__tarjanLowLink = { }
        for v in self.__nodes:
            self.__tarjanIndex[v] = None
        roots = self.roots()
        if (0 == len(roots)) and (0 < len(self.__nodes)):
            raise Exception('TARJAN: No roots found in graph with %d nodes' % (len(self.__nodes),))
        for r in roots:
            self._tarjan(r)
        self.__didTarjan = True

    def _tarjan (self, v):
        """Do the work of Tarjan's algorithm for a given root node."""
        if self.__tarjanIndex.get(v) is not None:
            # "Root" was already reached.
            return
        self.__tarjanIndex[v] = self.__tarjanLowLink[v] = self.__index
        self.__index += 1
        self.__stack.append(v)
        source = v
        for target in self.__edgeMap.get(source, []):
            if self.__tarjanIndex[target] is None:
                self._tarjan(target)
                self.__tarjanLowLink[v] = min(self.__tarjanLowLink[v], self.__tarjanLowLink[target])
            elif target in self.__stack:
                self.__tarjanLowLink[v] = min(self.__tarjanLowLink[v], self.__tarjanLowLink[target])
            else:
                pass

        if self.__tarjanLowLink[v] == self.__tarjanIndex[v]:
            scc = []
            while True:
                scc.append(self.__stack.pop())
                if v == scc[-1]:
                    break
            self.__sccOrder.append(scc)
            if 1 < len(scc):
                self.__scc.append(scc)
                [ self.__sccMap.setdefault(_v, scc) for _v in scc ]

    def scc (self, reset=False):
        """Return the strongly-connected components of the graph.

        The data structure is a set, each element of which is itself a
        set containing one or more nodes from the graph.

        @see: L{tarjan}.
        """
        if reset or (self.__scc is None):
            self.tarjan(reset)
        return self.__scc
    __scc = None

    def sccMap (self, reset=False):
        """Return a map from nodes to the strongly-connected component
        to which the node belongs.

        @keyword reset: If C{True}, the L{tarjan} method will be
        re-invoked, propagating the C{reset} value.  If C{False}
        (default), a cached value will be returned if available.

        @see: L{tarjan}.
        """
        if reset or (self.__sccMap is None):
            self.tarjan(reset)
        return self.__sccMap
    __sccMap = None

    def sccOrder (self, reset=False):
        """Return the strongly-connected components in order.

        The data structure is a list, in dependency order, of strongly
        connected components (which can be single nodes).  Appearance
        of a node in a set earlier in the list indicates that it has
        no dependencies on any node that appears in a subsequent set.
        This order is preferred over L{dfsOrder} for code generation,
        since it detects loops.

        @see: L{tarjan}.
        """
        if reset or (self.__sccOrder is None):
            self.tarjan(reset)
        return self.__sccOrder
    __sccOrder = None

    def sccForNode (self, node, **kw):
        """Return the strongly-connected component to which the given
        node belongs.

        Any keywords suppliend when invoking this method are passed to
        the L{sccMap} method.

        @return: The SCC set, or C{None} if the node is not present in
        the results of Tarjan's algorithm."""

        return self.sccMap(**kw).get(node)

    def cyclomaticComplexity (self):
        """Return the cyclomatic complexity of the graph."""
        self.tarjan()
        return len(self.__edges) - len(self.__nodes) + 2 * len(self.__scc)

    def __dfsWalk (self, source):
        assert not (source in self.__dfsWalked)
        self.__dfsWalked.add(source)
        for target in self.__edgeMap.get(source, []):
            if not (target in self.__dfsWalked):
                self.__dfsWalk(target)
        self.__dfsOrder.append(source)

    def _generateDOT (self, title='UNKNOWN', labeller=None):
        node_map = { }
        idx = 1
        for n in self.__nodes:
            node_map[n] = idx
            idx += 1
        text = []
        text.append('digraph "%s" {' % (title,))
        for n in self.__nodes:
            if labeller is not None:
                nn = labeller(n)
            else:
                nn = str(n)
            text.append('%s [shape=box,label="%s"];' % (node_map[n], nn))
        for s in self.__nodes:
            for d in self.__edgeMap.get(s, []):
                if s != d:
                    text.append('%s -> %s;' % (node_map[s], node_map[d]))
        text.append("};")
        return "\n".join(text)

    def dfsOrder (self, reset=False):
        """Return the nodes of the graph in U{depth-first-search
        order<http://en.wikipedia.org/wiki/Depth-first_search>}.

        The data structure is a list.  Calculated lists are retained
        and returned on future invocations, subject to the C{reset}
        keyword.

        @keyword reset: If C{True}, discard cached results and recompute the order."""
        if reset or (self.__dfsOrder is None):
            self.__dfsWalked = set()
            self.__dfsOrder = []
            for root in self.roots(reset=reset):
                self.__dfsWalk(root)
            self.__dfsWalked = None
            if len(self.__dfsOrder) != len(self.__nodes):
                raise Exception('DFS walk did not cover all nodes (walk %d versus nodes %d)' % (len(self.__dfsOrder), len(self.__nodes)))
        return self.__dfsOrder

    def rootSetOrder (self):
        """Return the nodes of the graph as a sequence of root sets.

        The first root set is the set of nodes that are roots: i.e.,
        have no incoming edges.  The second root set is the set of
        nodes that have incoming nodes in the first root set.  This
        continues until all nodes have been reached.  The sets impose
        a partial order on the nodes, without being as constraining as
        L{sccOrder}.

        @return: a list of the root sets."""
        order = []
        nodes = set(self.__nodes)
        edge_map = {}
        for (d, srcs) in six.iteritems(self.__edgeMap):
            edge_map[d] = srcs.copy()
        while nodes:
            freeset = set()
            for n in nodes:
                if not (n in edge_map):
                    freeset.add(n)
            if 0 == len(freeset):
                _log.error('dependency cycle in named components')
                return None
            order.append(freeset)
            nodes.difference_update(freeset)
            new_edge_map = {}
            for (d, srcs) in six.iteritems(edge_map):
                srcs.difference_update(freeset)
                if 0 != len(srcs):
                    new_edge_map[d] = srcs
            edge_map = new_edge_map
        return order

LocationPrefixRewriteMap_ = { }

def SetLocationPrefixRewriteMap (prefix_map):
    """Set the map that is used to by L{NormalizeLocation} to rewrite URI prefixes."""

    LocationPrefixRewriteMap_.clear()
    LocationPrefixRewriteMap_.update(prefix_map)

def NormalizeLocation (uri, parent_uri=None, prefix_map=None):
    """Normalize a URI against an optional parent_uri in the way that is
    done for C{schemaLocation} attribute values.

    If no URI schema is present, this will normalize a file system
    path.

    Optionally, the resulting absolute URI can subsequently be
    rewritten to replace specified prefix strings with alternative
    strings, e.g. to convert a remote URI to a local repository.  This
    rewriting is done after the conversion to an absolute URI, but
    before normalizing file system URIs.

    @param uri : The URI to normalize.  If C{None}, function returns
    C{None}

    @param parent_uri : The base URI against which normalization is
    done, if C{uri} is a relative URI.

    @param prefix_map : A map used to rewrite URI prefixes.  If
    C{None}, the value defaults to that stored by
    L{SetLocationPrefixRewriteMap}.

    """
    if uri is None:
        return uri
    if parent_uri is None:
        abs_uri = uri
    else:
        abs_uri = urlparse.urljoin(parent_uri, uri)
    if prefix_map is None:
        prefix_map = LocationPrefixRewriteMap_
    for (pfx, sub) in six.iteritems(prefix_map):
        if abs_uri.startswith(pfx):
            abs_uri = sub + abs_uri[len(pfx):]
    if 0 > abs_uri.find(':'):
        abs_uri = os.path.realpath(abs_uri)
    return abs_uri


def DataFromURI (uri, archive_directory=None):
    """Retrieve the contents of the uri as raw data.

    If the uri does not include a scheme (e.g., C{http:}), it is
    assumed to be a file path on the local system."""

    from pyxb.utils.six.moves.urllib.request import urlopen
    stream = None
    exc = None
    # Only something that has a colon is a non-file URI.  Some things
    # that have a colon are a file URI (sans schema).  Prefer urllib2,
    # but allow urllib (which apparently works better on Windows).
    if 0 <= uri.find(':'):
        try:
            stream = urlopen(uri)
        except Exception as e:
            exc = e
        if (stream is None) and six.PY2:
            import urllib
            try:
                stream = urllib.urlopen(uri)
                exc = None
            except:
                # Prefer urllib exception
                pass
    if stream is None:
        # No go as URI; give file a chance
        try:
            stream = open(uri, 'rb')
            exc = None
        except Exception as e:
            if exc is None:
                exc = e
    if exc is not None:
        _log.error('open %s', uri, exc_info=exc)
        raise exc
    try:
        # Protect this in case whatever stream is doesn't have an fp
        # attribute.
        if isinstance(stream, six.moves.file) or isinstance(stream.fp, six.moves.file):
            archive_directory = None
    except:
        pass
    xmld = stream.read()
    if archive_directory:
        base_name = os.path.basename(os.path.normpath(urlparse.urlparse(uri)[2]))
        counter = 1
        dest_file = os.path.join(archive_directory, base_name)
        while os.path.isfile(dest_file):
            dest_file = os.path.join(archive_directory, '%s.%d' % (base_name, counter))
            counter += 1
        try:
            OpenOrCreate(dest_file).write(xmld)
        except OSError as e:
            _log.warning('Unable to save %s in %s: %s', uri, dest_file, e)
    return xmld

def OpenOrCreate (file_name, tag=None, preserve_contents=False):
    """Return a file object used to write binary data into the given file.

    Use the C{tag} keyword to preserve the contents of existing files
    that are not supposed to be overwritten.

    To get a writable file but leaving any existing contents in place,
    set the C{preserve_contents} keyword to C{True}.  Normally, existing file
    contents are erased.

    The returned file pointer is positioned at the end of the file.

    @keyword tag: If not C{None} and the file already exists, absence
    of the given value in the first 4096 bytes of the file (decoded as
    UTF-8) causes an C{IOError} to be raised with C{errno} set to
    C{EEXIST}.  I.e., only files with this value in the first 4KB will
    be returned for writing.

    @keyword preserve_contents: This value controls whether existing
    contents of the file will be erased (C{False}, default) or left in
    place (C{True}).
    """
    (path, leaf) = os.path.split(file_name)
    if path:
        try:
            os.makedirs(path)
        except Exception as e:
            if not (isinstance(e, (OSError, IOError)) and (errno.EEXIST == e.errno)):
                raise
    fp = open(file_name, 'ab+')
    if (tag is not None) and (0 < os.fstat(fp.fileno()).st_size):
        fp.seek(0) # os.SEEK_SET
        blockd = fp.read(4096)
        blockt = blockd.decode('utf-8')
        if 0 > blockt.find(tag):
            raise OSError(errno.EEXIST, os.strerror(errno.EEXIST))
    if not preserve_contents:
        fp.seek(0) # os.SEEK_SET
        fp.truncate()
    else:
        fp.seek(2) # os.SEEK_END
    return fp

# hashlib didn't show up until 2.5, and sha is deprecated in 2.6.
__Hasher = None
try:
    import hashlib
    __Hasher = hashlib.sha1
except ImportError:
    import sha
    __Hasher = sha.new

def HashForText (text):
    """Calculate a cryptographic hash of the given string.

    For example, this is used to verify that a given module file
    contains bindings from a previous generation run for the same
    namespace.  See L{OpenOrCreate}.  If the text is in Unicode, the
    hash is calculated on the UTF-8 encoding of the text.

    @return: A C{str}, generally a sequence of hexadecimal "digit"s.
    """
    if isinstance(text, six.text_type):
        text = text.encode('utf-8')
    return __Hasher(text).hexdigest()

# uuid didn't show up until 2.5
__HaveUUID = False
try:
    import uuid
    __HaveUUID = True
except ImportError:
    import random
def _NewUUIDString ():
    """Obtain a UUID using the best available method.  On a version of
    python that does not incorporate the C{uuid} class, this creates a
    string combining the current date and time (to the second) with a
    random number.

    @rtype: C{str}
    """
    if __HaveUUID:
        return uuid.uuid1().urn
    return '%s:%08.8x' % (time.strftime('%Y%m%d%H%M%S'), random.randint(0, 0xFFFFFFFF))

class UniqueIdentifier (object):
    """Records a unique identifier, generally associated with a
    binding generation action.

    The identifier is a string, but gets wrapped in an instance of
    this class to optimize comparisons and reduce memory footprint.

    Invoking the constructor for this class on the same string
    multiple times will return the same Python object.

    An instance of this class compares equal to, and hashes equivalent
    to, the uid string.  When C{str}'d, the result is the uid; when
    C{repr}'d, the result is a constructor call to
    C{pyxb.utils.utility.UniqueIdentifier}.
    """

    # A map from UID string to the instance that represents it
    __ExistingUIDs = {}

    def uid (self):
        """The string unique identifier"""
        return self.__uid
    __uid = None

    # Support pickling, which is done using only the UID.
    def __getnewargs__ (self):
        return (self.__uid,)

    def __getstate__ (self):
        return self.__uid

    def __setstate__ (self, state):
        assert self.__uid == state

    # Singleton-like
    def __new__ (cls, *args):
        if 0 == len(args):
            uid = _NewUUIDString()
        else:
            uid = args[0]
        if isinstance(uid, UniqueIdentifier):
            uid = uid.uid()
        if not isinstance(uid, six.string_types):
            raise TypeError('UniqueIdentifier uid must be a string')
        rv = cls.__ExistingUIDs.get(uid)
        if rv is None:
            rv = super(UniqueIdentifier, cls).__new__(cls)
            rv.__uid = uid
            cls.__ExistingUIDs[uid] = rv
        return rv

    def associateObject (self, obj):
        """Associate the given object witth this identifier.

        This is a one-way association: the object is not provided with
        a return path to this identifier instance."""
        self.__associatedObjects.add(obj)
    def associatedObjects (self):
        """The set of objects that have been associated with this
        identifier instance."""
        return self.__associatedObjects
    __associatedObjects = None

    def __init__ (self, uid=None):
        """Create a new UniqueIdentifier instance.

        @param uid: The unique identifier string.  If present, it is
        the callers responsibility to ensure the value is universally
        unique.  If C{None}, one will be provided.
        @type uid: C{str} or C{unicode}
        """
        assert (uid is None) or (self.uid() == uid), 'UniqueIdentifier: ctor %s, actual %s' % (uid, self.uid())
        self.__associatedObjects = set()

    def __eq__ (self, other):
        if other is None:
            return False
        elif isinstance(other, UniqueIdentifier):
            other_uid = other.uid()
        elif isinstance(other, six.string_types):
            other_uid = other
        else:
            raise TypeError('UniqueIdentifier: Cannot compare with type %s' % (type(other),))
        return self.uid() == other_uid

    def __hash__ (self):
        return hash(self.uid())

    def __str__ (self):
        return self.uid()

    def __repr__ (self):
        return 'pyxb.utils.utility.UniqueIdentifier(%s)' % (repr(self.uid()),)

@BackfillComparisons
class UTCOffsetTimeZone (datetime.tzinfo):
    """A C{datetime.tzinfo} subclass that helps deal with UTC
    conversions in an ISO8601 world.

    This class only supports fixed offsets from UTC.
    """

    # Regular expression that matches valid ISO8601 time zone suffixes
    __Lexical_re = re.compile('^([-+])(\d\d):(\d\d)$')

    # The offset in minutes east of UTC.
    __utcOffset_min = 0

    # Same as __utcOffset_min, but as a datetime.timedelta
    __utcOffset_td = None

    # A zero-length duration
    __ZeroDuration = datetime.timedelta(0)

    # Range limits
    __MaxOffset_td = datetime.timedelta(hours=14)

    def __init__ (self, spec=None):
        """Create a time zone instance with a fixed offset from UTC.

        @param spec: Specifies the offset.  Can be an integer counting
        minutes east of UTC, the value C{None} (equal to 0 minutes
        east), or a string that conform to the ISO8601 time zone
        sequence (B{Z}, or B{[+-]HH:MM}).
        """

        if spec is not None:
            if isinstance(spec, six.string_types):
                if 'Z' == spec:
                    self.__utcOffset_min = 0
                else:
                    match = self.__Lexical_re.match(spec)
                    if match is None:
                        raise ValueError('Bad time zone: %s' % (spec,))
                    self.__utcOffset_min = int(match.group(2)) * 60 + int(match.group(3))
                    if '-' == match.group(1):
                        self.__utcOffset_min = - self.__utcOffset_min
            elif isinstance(spec, int):
                self.__utcOffset_min = spec
            elif isinstance(spec, datetime.timedelta):
                self.__utcOffset_min = spec.seconds // 60
            else:
                raise TypeError('%s: unexpected type %s' % (type(self), type(spec)))
        self.__utcOffset_td = datetime.timedelta(minutes=self.__utcOffset_min)
        if self.__utcOffset_td < -self.__MaxOffset_td or self.__utcOffset_td > self.__MaxOffset_td:
            raise ValueError('XSD timezone offset %s larger than %s' % (self.__utcOffset_td, self.__MaxOffset_td))
        if 0 == self.__utcOffset_min:
            self.__tzName = 'Z'
        elif 0 > self.__utcOffset_min:
            self.__tzName = '-%02d:%02d' % divmod(-self.__utcOffset_min, 60)
        else:
            self.__tzName = '+%02d:%02d' % divmod(self.__utcOffset_min, 60)

    def utcoffset (self, dt):
        """Returns the constant offset for this zone."""
        return self.__utcOffset_td

    def tzname (self, dt):
        """Return the name of the timezone in the format expected by XML Schema."""
        return self.__tzName

    def dst (self, dt):
        """Returns a constant zero duration."""
        return self.__ZeroDuration

    def __otherForComparison (self, other):
        if isinstance(other, UTCOffsetTimeZone):
            return other.__utcOffset_min
        return other.utcoffset(datetime.datetime.now())

    def __hash__ (self):
        return hash(self.__utcOffset_min)

    def __eq__ (self, other):
        return self.__utcOffset_min == self.__otherForComparison(other)

    def __lt__ (self, other):
        return self.__utcOffset_min < self.__otherForComparison(other)

class LocalTimeZone (datetime.tzinfo):
    """A C{datetime.tzinfo} subclass for the local time zone.

    Mostly pinched from the C{datetime.tzinfo} documentation in Python 2.5.1.
    """

    __STDOffset = datetime.timedelta(seconds=-time.timezone)
    __DSTOffset = __STDOffset
    if time.daylight:
        __DSTOffset = datetime.timedelta(seconds=-time.altzone)
    __ZeroDelta = datetime.timedelta(0)
    __DSTDelta = __DSTOffset - __STDOffset

    def utcoffset (self, dt):
        if self.__isDST(dt):
            return self.__DSTOffset
        return self.__STDOffset

    def dst (self, dt):
        if self.__isDST(dt):
            return self.__DSTDelta
        return self.__ZeroDelta

    def tzname (self, dt):
        return time.tzname[self.__isDST(dt)]

    def __isDST (self, dt):
        tt = (dt.year, dt.month, dt.day,
              dt.hour, dt.minute, dt.second,
              0, 0, -1)
        tt = time.localtime(time.mktime(tt))
        return tt.tm_isdst > 0

class PrivateTransient_mixin (pyxb.cscRoot):
    """Emulate the B{transient} keyword from Java for private member
    variables.

    This class defines a C{__getstate__} method which returns a copy
    of C{self.__dict__} with certain members removed.  Specifically,
    if a string "s" appears in a class member variable named
    C{__PrivateTransient} defined in the "Class" class, then the
    corresponding private variable "_Class__s" will be removed from
    the state dictionary.  This is used to eliminate unnecessary
    fields from instances placed in L{namespace
    archives<pyxb.namespace.archive.NamespaceArchive>} without having
    to implement a C{__getstate__} method in every class in the
    instance hierarchy.

    For an example, see
    L{pyxb.xmlschema.structures._SchemaComponent_mixin}

    If you use this, it is your responsibility to define the
    C{__PrivateTransient} class variable and add to it the required
    variable names.

    Classes that inherit from this are free to define their own
    C{__getstate__} method, which may or may not invoke the superclass
    one.  If you do this, be sure that the class defining
    C{__getstate__} lists L{PrivateTransient_mixin} as one of its
    direct superclasses, lest the latter end up earlier in the mro and
    consequently bypass the local override.
    """

    # Suffix used when creating the class member variable in which the
    # transient members are cached.
    __Attribute = '__PrivateTransient'

    def __getstate__ (self):
        state = self.__dict__.copy()
        # Note that the aggregate set is stored in a class variable
        # with a slightly different name than the class-level set.
        attr = '_%s%s_' % (self.__class__.__name__, self.__Attribute)
        skipped = getattr(self.__class__, attr, None)
        if skipped is None:
            skipped = set()
            for cl in self.__class__.mro():
                for (k, v) in six.iteritems(cl.__dict__):
                    if k.endswith(self.__Attribute):
                        cl2 = k[:-len(self.__Attribute)]
                        skipped.update([ '%s__%s' % (cl2, _n) for _n in v ])
            setattr(self.__class__, attr, skipped)
        for k in skipped:
            if state.get(k) is not None:
                del state[k]
        # Uncomment the following to test whether undesirable types
        # are being pickled, generally by accidently leaving a
        # reference to one in an instance private member.
        #for (k, v) in six.iteritems(state):
        #    import pyxb.namespace
        #    import xml.dom
        #    import pyxb.xmlschema.structures
        #    if isinstance(v, (pyxb.namespace.NamespaceContext, xml.dom.Node, pyxb.xmlschema.structures.Schema)):
        #        raise pyxb.LogicError('Unexpected instance of %s key %s in %s' % (type(v), k, self))

        return state

def GetMatchingFiles (path, pattern=None, default_path_wildcard=None, default_path=None, prefix_pattern=None, prefix_substituend=None):
    """Provide a list of absolute paths to files present in any of a
    set of directories and meeting certain criteria.

    This is used, for example, to locate namespace archive files
    within the archive path specified by the user.  One could use::

      files = GetMatchingFiles('&bundles//:+',
                               pattern=re.compile('.*\.wxs$'),
                               default_path_wildcard='+',
                               default_path='/usr/local/pyxb/nsarchives',
                               prefix_pattern='&',
                               prefix_substituend='/opt/pyxb')

    to obtain all files that can be recursively found within
    C{/opt/pyxb/bundles}, or non-recursively within
    C{/usr/local/pyxb/nsarchives}.

    @param path: A list of directories in which the search should be
    performed.  The entries are separated by os.pathsep, which is a
    colon on POSIX platforms and a semi-colon on Windows.  If a path
    entry ends with C{//} regardless of platform, the suffix C{//} is
    stripped and any directory beneath the path is scanned as well,
    recursively.

    @keyword pattern: Optional regular expression object used to
    determine whether a given directory entry should be returned.  If
    left as C{None}, all directory entries will be returned.

    @keyword default_path_wildcard: An optional string which, if
    present as a single directory in the path, is replaced by the
    value of C{default-path}.

    @keyword default_path: A system-defined directory which can be
    restored to the path by placing the C{default_path_wildcard} in
    the C{path}.

    @keyword prefix_pattern: An optional string which, if present at
    the start of a path element, is replaced by the value of
    C{prefix_substituend}.

    @keyword prefix_substituend: A system-defined string (path prefix)
    which can be combined with the user-provided path information to
    identify a file or subdirectory within an installation-specific
    area.
    """
    matching_files = []
    path_set = path.split(os.pathsep)
    while 0 < len(path_set):
        path = path_set.pop(0)
        if default_path_wildcard == path:
            if default_path is not None:
                path_set[0:0] = default_path.split(os.pathsep)
                default_path = None
            continue
        recursive = False
        if (prefix_pattern is not None) and path.startswith(prefix_pattern):
            path = os.path.join(prefix_substituend, path[len(prefix_pattern):])
        if path.endswith('//'):
            recursive = True
            path = path[:-2]
        if os.path.isfile(path):
            if (pattern is None) or (pattern.search(path) is not None):
                matching_files.append(path)
        else:
            for (root, dirs, files) in os.walk(path):
                for f in files:
                    if (pattern is None) or (pattern.search(f) is not None):
                        matching_files.append(os.path.join(root, f))
                if not recursive:
                    break
    return matching_files

@BackfillComparisons
class Location (object):
    __locationBase = None
    __lineNumber = None
    __columnNumber = None

    def __init__ (self, location_base=None, line_number=None, column_number=None):
        if isinstance(location_base, str):
            location_base = six.moves.intern(location_base)
        self.__locationBase = location_base
        self.__lineNumber = line_number
        self.__columnNumber = column_number

    def newLocation (self, locator=None, line_number=None, column_number=None):
        if locator is not None:
            try:
                line_number = locator.getLineNumber()
                column_number = locator.getColumnNumber()
            except:
                pass
        return Location(self.__locationBase, line_number, column_number)

    locationBase = property(lambda _s: _s.__locationBase)
    lineNumber = property(lambda _s: _s.__lineNumber)
    columnNumber = property(lambda _s: _s.__columnNumber)

    def __cmpSingleUnlessNone (self, v1, v2):
        if v1 is None:
            if v2 is None:
                return None
            return 1
        if v2 is None:
            return -1
        if v1 < v2:
            return -1
        if v1 == v2:
            return 0
        return 1

    def __cmpTupleUnlessNone (self, v1, v2):
        rv = self.__cmpSingleUnlessNone(v1.__locationBase, v2.__locationBase)
        if rv is None:
            rv = self.__cmpSingleUnlessNone(v1.__lineNumber, v2.__lineNumber)
        if rv is None:
            rv = self.__cmpSingleUnlessNone(v1.__columnNumber, v2.__columnNumber)
        return rv

    def __hash__ (self):
        return hash((self.__locationBase, self.__lineNumber, self.__columnNumber))

    def __eq__ (self, other):
        """Comparison by locationBase, then lineNumber, then columnNumber."""
        if other is None:
            return False
        rv = self.__cmpTupleUnlessNone(self, other)
        if rv is None:
            return True
        return 0 == rv

    def __lt__ (self, other):
        if other is None:
            return False
        rv = self.__cmpTupleUnlessNone(self, other)
        if rv is None:
            return False
        return -1 == rv

    def __str__ (self):
        if self.locationBase is None:
            lb = '<unknown>'
        else:
            # No, this should not be os.sep.  The location is
            # expected to be a URI.
            lb = self.locationBase.rsplit('/', 1)[-1]
        return '%s[%s:%s]' % (lb, self.lineNumber, self.columnNumber)

    def __repr__ (self):
        t = type(self)
        ctor = '%s.%s' % (t.__module__, t.__name__)
        return '%s(%s, %r, %r)' % (ctor, repr2to3(self.__locationBase), self.__lineNumber, self.__columnNumber)

class Locatable_mixin (pyxb.cscRoot):
    __location = None

    def __init__ (self, *args, **kw):
        self.__location = kw.pop('location', None)
        super(Locatable_mixin, self).__init__(*args, **kw)

    def _setLocation (self, location):
        self.__location = location

    def _location (self):
        return self.__location

def repr2to3 (v):
    """Filtered built-in repr for python 2/3 compatibility in
    generated bindings.

    All generated string values are to be unicode.  We always import
    unicode_literals from __future__, so we want plain quotes with no
    prefix u.  Strip that off.

    Integer constants should not have the suffix L even if they do not
    fit in a Python2 int.  The references generated through this
    function are never used for calculations, so the implicit cast to
    a larger type is sufficient.

    All other values use their standard representations.
    """
    if isinstance(v, six.string_types):
        qu = QuotedEscaped(v)
        if 'u' == qu[0]:
            return qu[1:]
        return qu
    if isinstance(v, six.integer_types):
        vs = repr(v)
        if vs.endswith('L'):
            return vs[:-1]
        return vs
    return repr(v)



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import unittest
from pyxb.utils.domutils import *
from xml.dom import Node
import xml.dom
import pyxb.namespace

def NonTextSibling (n):
    while n.TEXT_NODE == n.nodeType:
        n = n.nextSibling
    return n

class TestInScopeNames (unittest.TestCase):
    def show (self, node):
        xmlns_map = pyxb.namespace.NamespaceContext.GetNodeContext(node).inScopeNamespaces()
        #print '%s: %s' % (node.nodeName, ' ; '.join([ '%s=%s' % (_k, _v.uri()) for (_k, _v) in xmlns_map.items()]))
        xmlns_map = xmlns_map.copy()
        self.assertEqual(pyxb.namespace.XML.uri(), xmlns_map.pop('xml').uri())
        self.assertEqual(pyxb.namespace.XMLNamespaces.uri(), xmlns_map.pop('xmlns').uri())
        return xmlns_map

    def test_6_2_2 (self):
        xml = '''<?xml version="1.0"?>
<!-- initially, the default namespace is "books" -->
<book xmlns='urn:loc.gov:books'
      xmlns:isbn='urn:ISBN:0-395-36341-6'>
    <title>Cheaper by the Dozen</title>
    <isbn:number>1568491379</isbn:number>
    <notes>
      <p xmlns='http://www.w3.org/1999/xhtml'>
          This is a <i>funny</i> book!
      </p>
      <p>another graf without namespace change</p>
    </notes>
</book>'''
        book = StringToDOM(xml).documentElement
        self.assertEqual('book', book.localName)
        ns_ctx = pyxb.namespace.NamespaceContext.GetNodeContext(book)
        xmlns_map = self.show(book)
        self.assertEqual(2, len(xmlns_map))
        self.assertEqual('urn:loc.gov:books', xmlns_map[None].uri())
        self.assertEqual('urn:ISBN:0-395-36341-6', xmlns_map['isbn'].uri())

        title = NonTextSibling(book.firstChild)
        self.assertEqual('title', title.localName)
        xmlns_map =self.show(title)
        self.assertEqual(2, len(xmlns_map))
        self.assertEqual('urn:loc.gov:books', xmlns_map[None].uri())
        self.assertEqual('urn:ISBN:0-395-36341-6', xmlns_map['isbn'].uri())

        number = NonTextSibling(title.nextSibling)
        notes = NonTextSibling(number.nextSibling)
        while notes.TEXT_NODE == notes.nodeType:
            notes = notes.nextSibling
        self.assertEqual('notes', notes.localName)

        p = NonTextSibling(notes.firstChild)
        xmlns_map = self.show(p)
        self.assertEqual('p', p.localName)
        self.assertEqual(2, len(xmlns_map))
        self.assertEqual(None, xmlns_map.get('xsi'))
        self.assertEqual('http://www.w3.org/1999/xhtml', xmlns_map[None].uri())
        self.assertEqual('urn:ISBN:0-395-36341-6', xmlns_map['isbn'].uri())

        x = NonTextSibling(p.nextSibling)
        xmlns_map = self.show(x)
        self.assertEqual('p', x.localName)
        self.assertEqual(2, len(xmlns_map))
        self.assertEqual('urn:loc.gov:books', xmlns_map[None].uri())
        self.assertEqual('urn:ISBN:0-395-36341-6', xmlns_map['isbn'].uri())

    def test_6_2_3 (self):
        xml = '''<?xml version='1.0'?>
<Beers>
  <table xmlns='http://www.w3.org/1999/xhtml'>
   <th><td>Name</td><td>Origin</td><td>Description</td></th>
   <tr>
     <td><brandName xmlns="">Huntsman</brandName></td>
     <td><origin xmlns="">Bath, UK</origin></td>
     <td>
       <details xmlns=""><class>Bitter</class><hop>Fuggles</hop>
         <pro>Wonderful hop, light alcohol, good summer beer</pro>
         <con>Fragile; excessive variance pub to pub</con>
         </details>
        </td>
      </tr>
    </table>
  </Beers>'''
        Beers = StringToDOM(xml).documentElement
        ns_ctx = pyxb.namespace.NamespaceContext.GetNodeContext(Beers)
        xmlns_map = self.show(Beers)
        self.assertEqual(0, len(xmlns_map))
        table = NonTextSibling(Beers.firstChild)
        self.assertEqual('table', table.localName)
        xmlns_map = self.show(table)
        self.assertEqual(1, len(xmlns_map))
        self.assertEqual('http://www.w3.org/1999/xhtml', xmlns_map[None].uri())
        th = NonTextSibling(table.firstChild)
        self.assertEqual('th', th.localName)
        tr = NonTextSibling(th.nextSibling)
        self.assertEqual('tr', tr.localName)
        #brandName = table.firstChild.nextSibling.nextSibling.nextSibling.firstChild.nextSibling.nextSibling.nextSibling.firstChild
        td = NonTextSibling(tr.firstChild)
        self.assertEqual('td', td.localName)
        brandName = td.firstChild
        self.assertEqual('brandName', brandName.localName)
        xmlns_map = self.show(brandName)
        self.assertEqual(0, len(xmlns_map))


class TestNamespaceURIs (unittest.TestCase):
    # Make sure we agree with xml.dom on what the core namespace URIs are
    def testURI (self):
        self.assertTrue(xml.dom.EMPTY_NAMESPACE is None)
        self.assertEqual(xml.dom.XML_NAMESPACE, pyxb.namespace.XML.uri())
        self.assertEqual(xml.dom.XMLNS_NAMESPACE, pyxb.namespace.XMLNamespaces.uri())
        self.assertEqual(xml.dom.XHTML_NAMESPACE, pyxb.namespace.XHTML.uri())

if '__main__' == __name__:
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import print_function
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
from pyxb.exceptions_ import *
import unittest
import pyxb.binding.datatypes as xsd
import pyxb.namespace
xsi = pyxb.namespace.XMLSchema_instance

class Test_QName (unittest.TestCase):
    def testValid (self):
        dns = pyxb.namespace.CreateAbsentNamespace()
        nsc = pyxb.namespace.NamespaceContext(target_namespace=dns)
        nsc.declareNamespace(xsi, 'xsi')

        self.assertEqual('schema', xsd.QName('schema', _xmlns_context=nsc))
        self.assertEqual('schema', xsd.QName('schema'))
        self.assertEqual(xsi.createExpandedName('something'), xsd.QName('xsi:something', _xmlns_context=nsc))
        with self.assertRaises(pyxb.QNameResolutionError) as cm:
            xsd.QName('xs:something', _xmlns_context=nsc)
        self.assertEqual('xs:something', cm.exception.qname)
        self.assertEqual(nsc, cm.exception.namespaceContext)
        with self.assertRaises(pyxb.QNameResolutionError) as cm:
            xsd.QName('xsi:something')
        self.assertEqual('xsi:something', cm.exception.qname)
        self.assertEqual(None, cm.exception.namespaceContext)
        self.assertEqual('with.dots', xsd.QName('with.dots', _xmlns_context=nsc))

    def testInvalid (self):
        invalid = [ '-NonName', '-also:-not', 'and:-not', 'too:many:colons', ' whitespace ' ]
        for f in invalid:
            try:
                xsd.QName(f)
                print('Unexpected pass with %s' % (f,))
            except:
                pass
            self.assertRaises(SimpleTypeValueError, xsd.QName, f)

if __name__ == '__main__':
    unittest.main()



from pyxb.bundles.common.raw.xhtml1 import *
import pyxb.bundles.common.raw.xhtml1 as _raw

# The order of elements in an XHTML document is information that is
# not reflected in the schema information model.  For XHTML complex
# types replace the global validation configuration with one where the
# content list is strictly obeyed when generating documents.

DefaultValidationConfig = pyxb.GlobalValidationConfig.copy()
"""The validation configuration that applies to complex types in this namespace."""

DefaultValidationConfig._setContentInfluencesGeneration(DefaultValidationConfig.ALWAYS)
DefaultValidationConfig._setOrphanElementInContent(DefaultValidationConfig.RAISE_EXCEPTION)
DefaultValidationConfig._setInvalidElementInContent(DefaultValidationConfig.RAISE_EXCEPTION)

def _setValidationConfig ():
    import inspect
    import sys
    import pyxb.binding.basis

    for (n, v) in inspect.getmembers(_raw):
        if inspect.isclass(v) and issubclass(v, pyxb.binding.basis._TypeBinding_mixin):
            v._SetValidationConfig(DefaultValidationConfig)

_setValidationConfig()



# -*- coding: utf-8 -*-
# Copyright 2009-2013, Peter A. Bigot
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain a
# copy of the License at:
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Classes and global objects related to archiving U{XML
Namespaces<http://www.w3.org/TR/2006/REC-xml-names-20060816/index.html>}."""

import logging
import os
import os.path
import pyxb
import pyxb.utils.utility
from pyxb.utils import six

_log = logging.getLogger(__name__)

PathEnvironmentVariable = 'PYXB_ARCHIVE_PATH'
"""Environment variable from which default path to pre-loaded namespaces is
read.  The value should be a colon-separated list of absolute paths.  The
character C{&} at the start of a member of the list is replaced by the path to
the directory where the %{pyxb} modules are found, including a trailing C{/}.
For example, use C{&pyxb/bundles//} to enable search of any archive bundled
with PyXB.

@note: If you put a path separator between C{&} and the following path, this
will cause the substitution to be ignored."""

DefaultArchivePrefix = os.path.realpath(os.path.join(os.path.dirname( __file__), '../..'))
"""The default archive prefix, substituted for C{&} in C{PYXB_ARCHIVE_PATH}."""

def GetArchivePath ():
    """Return the archive path as defined by the L{PathEnvironmentVariable},
    or C{None} if that variable is not defined."""
    return os.environ.get(PathEnvironmentVariable)

# Stuff required for pickling
from pyxb.utils.six.moves import cPickle as pickle
import re

class NamespaceArchive (object):
    """Represent a file from which one or more namespaces can be read, or to
    which they will be written."""

    # A code used to identify the format of the archive, so we don't
    # mis-interpret its contents.
    # YYYYMMDDHHMM
    __PickleFormat = '200907190858'

    @classmethod
    def _AnonymousCategory (cls):
        """The category name to use when storing references to anonymous type
        definitions.  For example, attribute definitions defined within an
        attribute use in a model group definition.that can be referenced frojm
        ax different namespace."""
        return cls.__AnonymousCategory
    __AnonymousCategory = '_anonymousTypeDefinition'

    @classmethod
    def PicklingArchive (cls):
        """Return a reference to a set specifying the namespace instances that
        are being archived.

        This is needed to determine whether a component must be serialized as
        aa reference."""
        # NB: Use root class explicitly.  If we use cls, when this is invoked
        # by subclasses it gets mangled using the subclass name so the one
        # defined in this class is not found
        return NamespaceArchive.__PicklingArchive
    # Class variable recording the namespace that is currently being
    # pickled.  Used to prevent storing components that belong to
    # other namespaces.  Should be None unless within an invocation of
    # SaveToFile.
    __PicklingArchive = None

    __NamespaceArchives = None
    """A mapping from generation UID to NamespaceArchive instances."""

    def discard (self):
        """Remove this archive from the set of available archives.

        This is invoked when an archive contains a namespace that the user has
        specified should not be loaded."""
        del self.__NamespaceArchives[self.generationUID()]
        for ns in self.__namespaces:
            ns._removeArchive(self)

    @classmethod
    def __GetArchiveInstance (cls, archive_file, stage=None):
        """Return a L{NamespaceArchive} instance associated with the given file.

        To the extent possible, the same file accessed through different paths
        returns the same L{NamespaceArchive} instance.
        """

        nsa = NamespaceArchive(archive_path=archive_file, stage=cls._STAGE_uid)
        rv = cls.__NamespaceArchives.get(nsa.generationUID(), nsa)
        if rv == nsa:
            cls.__NamespaceArchives[rv.generationUID()] = rv
        rv._readToStage(stage)
        return rv

    __ArchivePattern_re = re.compile('\.wxs$')

    @classmethod
    def PreLoadArchives (cls, archive_path=None, reset=False):
        """Scan for available archives, associating them with namespaces.

        This only validates potential archive contents; it does not load
        namespace data from the archives.

        @keyword archive_path: A list of files or directories in which
        namespace archives can be found.  The entries are separated by
        os.pathsep, which is a colon on POSIX platforms and a semi-colon on
        Windows.  See L{PathEnvironmentVariable}.  Defaults to
        L{GetArchivePath()}.  If not defaulted, C{reset} will be forced to
        C{True}.  For any directory in the path, all files ending with
        C{.wxs} are examined.

        @keyword reset: If C{False} (default), the most recently read set of
        archives is returned; if C{True}, the archive path is re-scanned and the
        namespace associations validated.
        """

        from pyxb.namespace import builtin

        reset = reset or (archive_path is not None) or (cls.__NamespaceArchives is None)
        if reset:
            # Get a list of pre-existing archives, initializing the map if
            # this is the first time through.
            if cls.__NamespaceArchives is None:
                cls.__NamespaceArchives = { }
            existing_archives = set(six.itervalues(cls.__NamespaceArchives))
            archive_set = set()

            # Ensure we have an archive path.  If not, don't do anything.
            if archive_path is None:
                archive_path = GetArchivePath()
            if archive_path is not None:

                # Get archive instances for everything in the archive path
                candidate_files = pyxb.utils.utility.GetMatchingFiles(archive_path, cls.__ArchivePattern_re,
                                                                      default_path_wildcard='+', default_path=GetArchivePath(),
                                                                      prefix_pattern='&', prefix_substituend=DefaultArchivePrefix)
                for afn in candidate_files:
                    try:
                        nsa = cls.__GetArchiveInstance(afn, stage=cls._STAGE_readModules)
                        archive_set.add(nsa)
                    except pickle.UnpicklingError:
                        _log.exception('Cannot unpickle archive %s', afn)
                    except pyxb.NamespaceArchiveError:
                        _log.exception('Cannot process archive %s', afn)

                # Do this for two reasons: first, to get an iterable that won't
                # cause problems when we remove unresolvable archives from
                # archive_set; and second to aid with forced dependency inversion
                # testing
                ordered_archives = sorted(list(archive_set), key=lambda _a: _a.archivePath())
                ordered_archives.reverse()

                # Create a graph that identifies dependencies between the archives
                archive_map = { }
                for a in archive_set:
                    archive_map[a.generationUID()] = a
                archive_graph = pyxb.utils.utility.Graph()
                for a in ordered_archives:
                    prereqs = a._unsatisfiedModulePrerequisites()
                    if 0 < len(prereqs):
                        for p in prereqs:
                            if builtin.BuiltInObjectUID == p:
                                continue
                            da = archive_map.get(p)
                            if da is None:
                                _log.warning('%s depends on unavailable archive %s', a, p)
                                archive_set.remove(a)
                            else:
                                archive_graph.addEdge(a, da)
                    else:
                        archive_graph.addRoot(a)

                # Verify that there are no dependency loops.
                archive_scc = archive_graph.sccOrder()
                for scc in archive_scc:
                    if 1 < len(scc):
                        raise pyxb.LogicError("Cycle in archive dependencies.  How'd you do that?\n  " + "\n  ".join([ _a.archivePath() for _a in scc ]))
                    archive = scc[0]
                    if not (archive in archive_set):
                        archive.discard()
                        existing_archives.remove(archive)
                        continue
                    #archive._readToStage(cls._STAGE_COMPLETE)

            # Discard any archives that we used to know about but now aren't
            # supposed to.  @todo make this friendlier in the case of archives
            # we've already incorporated.
            for archive in existing_archives.difference(archive_set):
                _log.info('Discarding excluded archive %s', archive)
                archive.discard()

    def archivePath (self):
        """Path to the file in which this namespace archive is stored."""
        return self.__archivePath
    __archivePath = None

    def generationUID (self):
        """The unique identifier for the generation that produced this archive."""
        return self.__generationUID
    __generationUID = None

    def isLoadable (self):
        """Return C{True} iff it is permissible to load the archive.
        Archives created for output cannot be loaded."""
        return self.__isLoadable
    __isLoadable = None

    def __locateModuleRecords (self):
        self.__moduleRecords = set()
        namespaces = set()
        for ns in pyxb.namespace.utility.AvailableNamespaces():
            # @todo allow these; right now it's usually the XML
            # namespace and we're not prepared to reconcile
            # redefinitions of those components.
            if ns.isUndeclaredNamespace():
                continue
            mr = ns.lookupModuleRecordByUID(self.generationUID())
            if mr is not None:
                namespaces.add(ns)
                mr.prepareForArchive(self)
                self.__moduleRecords.add(mr)
        self.__namespaces.update(namespaces)
    def moduleRecords (self):
        """Return the set of L{module records <ModuleRecord>} stored in this
        archive.

        Each module record represents"""
        return self.__moduleRecords
    __moduleRecords = None

    @classmethod
    def ForPath (cls, archive_file):
        """Return the L{NamespaceArchive} instance that can be found at the
        given path."""
        return cls.__GetArchiveInstance(archive_file)

    # States in the finite automaton that is used to read archive contents.
    _STAGE_UNOPENED = 0         # Haven't even checked for existence
    _STAGE_uid = 1              # Verified archive exists, obtained generation UID from it
    _STAGE_readModules = 2      # Read module records from archive, which includes UID dependences
    _STAGE_validateModules = 3  # Verified pre-requisites for module loading
    _STAGE_readComponents = 4   # Extracted components from archive and integrated into namespaces
    _STAGE_COMPLETE = _STAGE_readComponents

    def _stage (self):
        return self.__stage
    __stage = None

    def __init__ (self, archive_path=None, generation_uid=None, loadable=True, stage=None):
        """Create a new namespace archive.

        If C{namespaces} is given, this is an output archive.

        If C{namespaces} is absent, this is an input archive.

        @raise IOError: error attempting to read the archive file
        @raise pickle.UnpicklingError: something is wrong with the format of the library
        """
        self.__namespaces = set()
        if generation_uid is not None:
            if archive_path:
                raise pyxb.LogicError('NamespaceArchive: cannot define both namespaces and archive_path')
            self.__generationUID = generation_uid
            self.__locateModuleRecords()
        elif archive_path is not None:
            if generation_uid is not None:
                raise pyxb.LogicError('NamespaceArchive: cannot provide generation_uid with archive_path')
            self.__archivePath = archive_path
            self.__stage = self._STAGE_UNOPENED
            self.__isLoadable = loadable
            if self.__isLoadable:
                if stage is None:
                    stage = self._STAGE_moduleRecords
                self._readToStage(stage)
        else:
            pass

    def add (self, namespace):
        """Add the given namespace to the set that is to be stored in this archive."""
        if namespace.isAbsentNamespace():
            raise pyxb.NamespaceArchiveError('Cannot archive absent namespace')
        self.__namespaces.add(namespace)

    def update (self, namespace_set):
        """Add the given namespaces to the set that is to be stored in this archive."""
        [ self.add(_ns) for _ns in namespace_set ]

    def namespaces (self):
        """Set of namespaces that can be read from this archive."""
        return self.__namespaces
    __namespaces = None

    def __createPickler (self, output):
        if isinstance(output, six.string_types):
            output = open(output, 'wb')
        pickler = pickle.Pickler(output, -1)

        # The format of the archive
        pickler.dump(NamespaceArchive.__PickleFormat)

        # The UID for the set
        assert self.generationUID() is not None
        pickler.dump(self.generationUID())

        return pickler

    def __createUnpickler (self):
        unpickler = pickle.Unpickler(open(self.__archivePath, 'rb'))

        fmt = unpickler.load()
        if self.__PickleFormat != fmt:
            raise pyxb.NamespaceArchiveError('Archive format is %s, require %s' % (fmt, self.__PickleFormat))

        self.__generationUID = unpickler.load()

        return unpickler

    def __readModules (self, unpickler):
        mrs = unpickler.load()
        assert isinstance(mrs, set), 'Expected set got %s from %s' % (type(mrs), self.archivePath())
        if self.__moduleRecords is None:
            for mr in mrs.copy():
                mr2 = mr.namespace().lookupModuleRecordByUID(mr.generationUID())
                if mr2 is not None:
                    mr2._setFromOther(mr, self)
                    mrs.remove(mr)
            self.__moduleRecords = set()
            assert 0 == len(self.__namespaces)
            for mr in mrs:
                mr._setArchive(self)
                ns = mr.namespace()
                ns.addModuleRecord(mr)
                self.__namespaces.add(ns)
                self.__moduleRecords.add(mr)
        else:
            # Verify the archive still has what was in it when we created this.
            for mr in mrs:
                mr2 = mr.namespace().lookupModuleRecordByUID(mr.generationUID())
                if not (mr2 in self.__moduleRecords):
                    raise pyxb.NamespaceArchiveError('Lost module record %s %s from %s' % (mr.namespace(), mr.generationUID(), self.archivePath()))

    def _unsatisfiedModulePrerequisites (self):
        prereq_uids = set()
        for mr in self.__moduleRecords:
            prereq_uids.update(mr.dependsOnExternal())
        return prereq_uids

    def __validatePrerequisites (self, stage):
        from pyxb.namespace import builtin
        prereq_uids = self._unsatisfiedModulePrerequisites()
        for uid in prereq_uids:
            if builtin.BuiltInObjectUID == uid:
                continue
            depends_on = self.__NamespaceArchives.get(uid)
            if depends_on is None:
                raise pyxb.NamespaceArchiveError('%s: archive depends on unavailable archive %s' % (self.archivePath(), uid))
            depends_on._readToStage(stage)

    def __validateModules (self):
        self.__validatePrerequisites(self._STAGE_validateModules)
        for mr in self.__moduleRecords:
            ns = mr.namespace()
            for base_uid in mr.dependsOnExternal():
                xmr = ns.lookupModuleRecordByUID(base_uid)
                if xmr is None:
                    raise pyxb.NamespaceArchiveError('Module %s depends on external module %s, not available in archive path' % (mr.generationUID(), base_uid))
                if not xmr.isIncorporated():
                    _log.info('Need to incorporate data from %s', xmr)
                else:
                    _log.info('Have required base data %s', xmr)

            for origin in mr.origins():
                for (cat, names) in six.iteritems(origin.categoryMembers()):
                    if not (cat in ns.categories()):
                        continue
                    cross_objects = names.intersection(six.iterkeys(ns.categoryMap(cat)))
                    if 0 < len(cross_objects):
                        raise pyxb.NamespaceArchiveError('Archive %s namespace %s module %s origin %s archive/active conflict on category %s: %s' % (self.__archivePath, ns, mr, origin, cat, " ".join(cross_objects)))
                    _log.info('%s no conflicts on %d names', cat, len(names))

    def __readComponentSet (self, unpickler):
        self.__validatePrerequisites(self._STAGE_readComponents)
        for n in range(len(self.__moduleRecords)):
            ns = unpickler.load()
            mr = ns.lookupModuleRecordByUID(self.generationUID())
            assert mr in self.__moduleRecords
            assert not mr.isIncorporated()
            objects = unpickler.load()
            mr._loadCategoryObjects(objects)

    __unpickler = None
    def _readToStage (self, stage):
        if self.__stage is None:
            raise pyxb.NamespaceArchiveError('Attempt to read from invalid archive %s' % (self,))
        try:
            while self.__stage < stage:
                if self.__stage < self._STAGE_uid:
                    self.__unpickler = self.__createUnpickler()
                    self.__stage = self._STAGE_uid
                    continue
                if self.__stage < self._STAGE_readModules:
                    assert self.__unpickler is not None
                    self.__readModules(self.__unpickler)
                    self.__stage = self._STAGE_readModules
                    continue
                if self.__stage < self._STAGE_validateModules:
                    self.__validateModules()
                    self.__stage = self._STAGE_validateModules
                    continue
                if self.__stage < self._STAGE_readComponents:
                    assert self.__unpickler is not None
                    self.__stage = self._STAGE_readComponents
                    self.__readComponentSet(self.__unpickler)
                    self.__unpickler = None
                    continue
                raise pyxb.LogicError('Too many stages (at %s, want %s)' % (self.__stage, stage))
        except:
            self.__stage = None
            self.__unpickler = None
            raise

    def readNamespaces (self):
        """Read all the components from this archive, integrating them into
        their respective namespaces."""
        self._readToStage(self._STAGE_COMPLETE)

    def writeNamespaces (self, output):
        """Store the namespaces into the archive.

        @param output: An instance substitutable for a writable file, or the
        name of a file to write to.
        """
        import sys

        assert NamespaceArchive.__PicklingArchive is None
        NamespaceArchive.__PicklingArchive = self
        assert self.__moduleRecords is not None

        # Recalculate the record/object associations: we didn't assign
        # anonymous names to the indeterminate scope objects because they
        # weren't needed for bindings, but they are needed in the archive.
        for mr in self.__moduleRecords:
            mr.namespace()._associateOrigins(mr)

        try:
            # See http://bugs.python.org/issue3338
            recursion_limit = sys.getrecursionlimit()
            sys.setrecursionlimit(10 * recursion_limit)

            pickler = self.__createPickler(output)

            assert isinstance(self.__moduleRecords, set)
            pickler.dump(self.__moduleRecords)

            for mr in self.__moduleRecords:
                pickler.dump(mr.namespace())
                pickler.dump(mr.categoryObjects())
        finally:
            sys.setrecursionlimit(recursion_limit)
        NamespaceArchive.__PicklingArchive = None

    def __str__ (self):
        archive_path = self.__archivePath
        if archive_path is None:
            archive_path = '??'
        return 'NSArchive@%s' % (archive_path,)

class _ArchivableObject_mixin (pyxb.cscRoot):
    """Mix-in to any object that can be stored in a namespace within an archive."""

    # Need to set this per category item
    __objectOrigin = None
    def _objectOrigin (self):
        return self.__objectOrigin
    def _setObjectOrigin (self, object_origin, override=False):
        if (self.__objectOrigin is not None) and (not override):
            if  self.__objectOrigin != object_origin:
                raise pyxb.LogicError('Inconsistent origins for object %s: %s %s' % (self, self.__objectOrigin, object_origin))
        else:
            self.__objectOrigin = object_origin

    def _prepareForArchive (self, archive):
        #assert self.__objectOrigin is not None
        if self._objectOrigin() is not None:
            return getattr(super(_ArchivableObject_mixin, self), '_prepareForArchive_csc', lambda *_args,**_kw: self)(self._objectOrigin().moduleRecord())
        assert not isinstance(self, pyxb.xmlschema.structures._NamedComponent_mixin)

    def _updateFromOther_csc (self, other):
        return getattr(super(_ArchivableObject_mixin, self), '_updateFromOther_csc', lambda *_args,**_kw: self)(other)

    def _updateFromOther (self, other):
        """Update this instance with additional information provided by the other instance.

        This is used, for example, when a built-in type is already registered
        in the namespace, but we've processed the corresponding schema and
        have obtained more details."""
        assert self != other
        return self._updateFromOther_csc(other)

    def _allowUpdateFromOther (self, other):
        from pyxb.namespace import builtin
        assert self._objectOrigin()
        return builtin.BuiltInObjectUID == self._objectOrigin().generationUID()

class _NamespaceArchivable_mixin (pyxb.cscRoot):
    """Encapsulate the operations and data relevant to archiving namespaces.

    This class mixes-in to L{pyxb.namespace.Namespace}"""

    def _reset (self):
        """CSC extension to reset fields of a Namespace.

        This one handles category-related data."""
        getattr(super(_NamespaceArchivable_mixin, self), '_reset', lambda *args, **kw: None)()
        self.__loadedFromArchive = None
        self.__wroteToArchive = None
        self.__active = False
        self.__moduleRecordMap = {}

    def _loadedFromArchive (self):
        return self.__loadedFromArchive

    __wroteToArchive = None
    __loadedFromArchive = None

    def isActive (self, empty_inactive=False):
        if self.__isActive and empty_inactive:
            for (ct, cm) in six.iteritems(self._categoryMap()):
                if 0 < len(cm):
                    return True
            return False
        return self.__isActive

    def _activate (self):
        self.__isActive = True
    __isActive = None

    def __init__ (self, *args, **kw):
        super(_NamespaceArchivable_mixin, self).__init__(*args, **kw)

    def _setLoadedFromArchive (self, archive):
        self.__loadedFromArchive = archive
        self._activate()
    def _setWroteToArchive (self, archive):
        self.__wroteToArchive = archive

    def _removeArchive (self, archive):
        # Yes, I do want this to raise KeyError if the archive is not present
        mr = self.__moduleRecordMap[archive.generationUID()]
        assert not mr.isIncorporated(), 'Removing archive %s after incorporation' % (archive.archivePath(),)
        del self.__moduleRecordMap[archive.generationUID()]

    def isLoadable (self):
        """Return C{True} iff the component model for this namespace can be
        loaded from a namespace archive."""
        for mr in self.moduleRecords():
            if mr.isLoadable():
                return True
        return False

    def isImportAugmentable (self):
        """Return C{True} iff the component model for this namespace may be
        extended by import directives.

        This is the case if the namespace has been marked with
        L{setImportAugmentable}, or if there is no archive or built-in that
        provides a component model for the namespace."""
        if self.__isImportAugmentable:
            return True
        for mr in self.moduleRecords():
            if mr.isLoadable() or mr.isIncorporated():
                return False
        return True
    def setImportAugmentable (self, value=True):
        self.__isImportAugmentable = value
    __isImportAugmentable = False

    def loadableFrom (self):
        """Return the list of archives from which components for this
        namespace can be loaded."""
        rv = []
        for mr in self.moduleRecords():
            if mr.isLoadable():
                rv.append(mr.archive())
        return rv

    def moduleRecords (self):
        return list(six.itervalues(self.__moduleRecordMap))
    __moduleRecordMap = None

    def addModuleRecord (self, module_record):
        assert isinstance(module_record, ModuleRecord)
        assert not (module_record.generationUID() in self.__moduleRecordMap)
        self.__moduleRecordMap[module_record.generationUID()] = module_record
        return module_record
    def lookupModuleRecordByUID (self, generation_uid, create_if_missing=False, *args, **kw):
        rv = self.__moduleRecordMap.get(generation_uid)
        if (rv is None) and create_if_missing:
            rv = self.addModuleRecord(ModuleRecord(self, generation_uid, *args, **kw))
        return rv

    def _setState_csc (self, kw):
        #assert not self.__isActive, 'ERROR: State set for active namespace %s' % (self,)
        return getattr(super(_NamespaceArchivable_mixin, self), '_getState_csc', lambda _kw: _kw)(kw)

    def markNotLoadable (self):
        """Prevent loading this namespace from an archive.

        This marks all archives in which the namespace appears, whether
        publically or privately, as not loadable."""
        if self._loadedFromArchive():
            raise pyxb.NamespaceError(self, 'cannot mark not loadable when already loaded')
        for mr in self.moduleRecords():
            mr._setIsLoadable(False)

class ModuleRecord (pyxb.utils.utility.PrivateTransient_mixin):
    __PrivateTransient = set()

    def namespace (self):
        return self.__namespace
    __namespace = None

    def archive (self):
        return self.__archive
    def _setArchive (self, archive):
        self.__archive = archive
        return self
    __archive = None
    __PrivateTransient.add('archive')

    def isPublic (self):
        return self.__isPublic
    def _setIsPublic (self, is_public):
        self.__isPublic = is_public
        return self
    __isPublic = None

    def isIncorporated (self):
        return self.__isIncorporated or (self.archive() is None)
    def markIncorporated (self):
        assert self.__isLoadable
        self.__isIncorporated = True
        self.__isLoadable = False
        return self
    __isIncorporated = None
    __PrivateTransient.add('isIncorporated')

    def isLoadable (self):
        return self.__isLoadable and (self.archive() is not None)
    def _setIsLoadable (self, is_loadable):
        self.__isLoadable = is_loadable
        return self
    __isLoadable = None

    def generationUID (self):
        return self.__generationUID
    __generationUID = None

    def origins (self):
        return list(six.itervalues(self.__originMap))
    def addOrigin (self, origin):
        assert isinstance(origin, _ObjectOrigin)
        assert not (origin.signature() in self.__originMap)
        self.__originMap[origin.signature()] = origin
        return origin
    def lookupOriginBySignature (self, signature):
        return self.__originMap.get(signature)
    def _setOrigins (self, origins):
        if self.__originMap is None:
            self.__originMap = {}
        else:
            self.__originMap.clear()
        [ self.addOrigin(_o) for _o in origins ]
        return self
    __originMap = None

    def hasMatchingOrigin (self, **kw):
        for origin in self.origins():
            if origin.match(**kw):
                return True
        return False

    def modulePath (self):
        return self.__modulePath
    def setModulePath (self, module_path):
        if isinstance(module_path, six.string_types):
            self.__modulePath = '.'.join(map(pyxb.utils.utility.MakeModuleElement, module_path.split('.')))
        else:
            assert (module_path is None)
            self.__modulePath = module_path
        return self
    __modulePath = None

    def referencedNamespaces (self):
        return self.__referencedNamespaces
    def _setReferencedNamespaces (self, referenced_namespaces):
        self.__referencedNamespaces.update(referenced_namespaces)
        return self
    def referenceNamespace (self, namespace):
        self.__referencedNamespaces.add(namespace)
        return namespace
    __referencedNamespaces = None

    __constructedLocally = False
    __PrivateTransient.add('constructedLocally')

    def __init__ (self, namespace, generation_uid, **kw):
        from pyxb.namespace import builtin

        super(ModuleRecord, self).__init__()
        self.__namespace = namespace
        assert (generation_uid != builtin.BuiltInObjectUID) or namespace.isBuiltinNamespace()
        self.__isPublic = kw.get('is_public', False)
        self.__isIncorporated = kw.get('is_incorporated', False)
        self.__isLoadable = kw.get('is_loadable', True)
        assert isinstance(generation_uid, pyxb.utils.utility.UniqueIdentifier)
        self.__generationUID = generation_uid
        self.__modulePath = kw.get('module_path')
        self.__originMap = {}
        self.__referencedNamespaces = set()
        self.__categoryObjects = { }
        self.__constructedLocally = True
        self.__dependsOnExternal = set()

    def _setFromOther (self, other, archive):
        if (not self.__constructedLocally) or other.__constructedLocally:
            raise pyxb.ImplementationError('Module record update requires local to be updated from archive')
        assert self.__generationUID == other.__generationUID
        assert self.__archive is None
        self.__isPublic = other.__isPublic
        assert not self.__isIncorporated
        self.__isLoadable = other.__isLoadable
        self.__modulePath = other.__modulePath
        self.__originMap.update(other.__originMap)
        self.__referencedNamespaces.update(other.__referencedNamespaces)
        if not (other.__categoryObjects is None):
            self.__categoryObjects.update(other.__categoryObjects)
        self.__dependsOnExternal.update(other.__dependsOnExternal)
        self._setArchive(archive)

    def categoryObjects (self):
        return self.__categoryObjects
    def resetCategoryObjects (self):
        self.__categoryObjects.clear()
        for origin in self.origins():
            origin.resetCategoryMembers()
    def _addCategoryObject (self, category, name, obj):
        obj._prepareForArchive(self)
        self.__categoryObjects.setdefault(category, {})[name] = obj
    def _loadCategoryObjects (self, category_objects):
        assert self.__categoryObjects is None
        assert not self.__constructedLocally
        ns = self.namespace()
        ns.configureCategories(six.iterkeys(category_objects))
        for (cat, obj_map) in six.iteritems(category_objects):
            current_map = ns.categoryMap(cat)
            for (local_name, component) in six.iteritems(obj_map):
                existing_component = current_map.get(local_name)
                if existing_component is None:
                    current_map[local_name] = component
                elif existing_component._allowUpdateFromOther(component):
                    existing_component._updateFromOther(component)
                else:
                    raise pyxb.NamespaceError(self, 'Load attempted to override %s %s in %s' % (cat, local_name, self.namespace()))
        self.markIncorporated()
    __categoryObjects = None
    __PrivateTransient.add('categoryObjects')

    def dependsOnExternal (self):
        return self.__dependsOnExternal
    __dependsOnExternal = None

    def prepareForArchive (self, archive):
        assert self.archive() is None
        self._setArchive(archive)
        ns = self.namespace()
        self.__dependsOnExternal.clear()
        for mr in ns.moduleRecords():
            if mr != self:
                _log.info('This gen depends on %s', mr)
                self.__dependsOnExternal.add(mr.generationUID())
        for obj in ns._namedObjects().union(ns.components()):
            if isinstance(obj, _ArchivableObject_mixin):
                if obj._objectOrigin():
                    obj._prepareForArchive(self)

    def completeGenerationAssociations (self):
        self.namespace()._transferReferencedNamespaces(self)
        self.namespace()._associateOrigins(self)

    def __str__ (self):
        return 'MR[%s]@%s' % (self.generationUID(), self.namespace())

class _ObjectOrigin (pyxb.utils.utility.PrivateTransient_mixin, pyxb.cscRoot):
    """Marker class for objects that can serve as an origin for an object in a
    namespace."""
    __PrivateTransient = set()

    def signature (self):
        return self.__signature
    __signature = None

    def moduleRecord (self):
        return self.__moduleRecord
    __moduleRecord = None

    def namespace (self):
        return self.moduleRecord().namespace()

    def generationUID (self):
        return self.moduleRecord().generationUID()

    def __init__ (self, namespace, generation_uid, **kw):
        self.__signature = kw.pop('signature', None)
        super(_ObjectOrigin, self).__init__(**kw)
        self.__moduleRecord = namespace.lookupModuleRecordByUID(generation_uid, create_if_missing=True, **kw)
        self.__moduleRecord.addOrigin(self)
        self.__categoryMembers = { }
        self.__categoryObjectMap = { }

    def resetCategoryMembers (self):
        self.__categoryMembers.clear()
        self.__categoryObjectMap.clear()
        self.__originatedComponents = None
    def addCategoryMember (self, category, name, obj):
        self.__categoryMembers.setdefault(category, set()).add(name)
        self.__categoryObjectMap.setdefault(category, {})[name] = obj
        self.__moduleRecord._addCategoryObject(category, name, obj)
    def categoryMembers (self):
        return self.__categoryMembers
    def originatedObjects (self):
        if self.__originatedObjects is None:
            components = set()
            [ components.update(six.itervalues(_v)) for _v in six.itervalues(self.__categoryObjectMap) ]
            self.__originatedObjects = frozenset(components)
        return self.__originatedObjects

    # The set of category names associated with objects.  Don't throw this
    # away and use categoryObjectMap.keys() instead: that's transient, and we
    # need this to have a value when read from an archive.
    __categoryMembers = None

    # Map from category name to a map from an object name to the object
    __categoryObjectMap = None
    __PrivateTransient.add('categoryObjectMap')

    # The set of objects that originated at this origin
    __originatedObjects = None
    __PrivateTransient.add('originatedObjects')

class _SchemaOrigin (_ObjectOrigin):
    """Holds the data regarding components derived from a single schema.

    Coupled to a particular namespace through the
    L{_NamespaceComponentAssociation_mixin}.
    """

    __PrivateTransient = set()

    def __setDefaultKW (self, kw):
        schema = kw.get('schema')
        if schema is not None:
            assert not ('location' in kw)
            kw['location'] = schema.location()
            assert not ('signature' in kw)
            kw['signature'] = schema.signature()
            assert not ('generation_uid' in kw)
            kw['generation_uid'] = schema.generationUID()
            assert not ('namespace' in kw)
            kw['namespace'] = schema.targetNamespace()
            assert not ('version' in kw)
            kw['version'] = schema.schemaAttribute('version')

    def match (self, **kw):
        """Determine whether this record matches the parameters.

        @keyword schema: a L{pyxb.xmlschema.structures.Schema} instance from
        which the other parameters are obtained.
        @keyword location: a schema location (URI)
        @keyword signature: a schema signature
        @return: C{True} iff I{either} C{location} or C{signature} matches."""
        self.__setDefaultKW(kw)
        location = kw.get('location')
        if (location is not None) and (self.location() == location):
            return True
        signature = kw.get('signature')
        if (signature is not None) and (self.signature() == signature):
            return True
        return False

    def location (self):
        return self.__location
    __location = None

    def schema (self):
        return self.__schema
    __schema = None
    __PrivateTransient.add('schema')

    def version (self):
        return self.__version
    __version = None

    def __init__ (self, **kw):
        self.__setDefaultKW(kw)
        self.__schema = kw.pop('schema', None)
        self.__location = kw.pop('location', None)
        self.__version = kw.pop('version', None)
        super(_SchemaOrigin, self).__init__(kw.pop('namespace'), kw.pop('generation_uid'), **kw)

    def __str__ (self):
        rv = [ '_SchemaOrigin(%s@%s' % (self.namespace(), self.location()) ]
        if self.version() is not None:
            rv.append(',version=%s' % (self.version(),))
        rv.append(')')
        return ''.join(rv)

class NamespaceDependencies (object):

    def rootNamespaces (self):
        return self.__rootNamespaces
    __rootNamespaces = None

    def namespaceGraph (self, reset=False):
        if reset or (self.__namespaceGraph is None):
            self.__namespaceGraph = pyxb.utils.utility.Graph()
            for ns in self.rootNamespaces():
                self.__namespaceGraph.addRoot(ns)

            # Make sure all referenced namespaces have valid components
            need_check = self.__rootNamespaces.copy()
            done_check = set()
            while 0 < len(need_check):
                ns = need_check.pop()
                ns.validateComponentModel()
                self.__namespaceGraph.addNode(ns)
                for rns in ns.referencedNamespaces().union(ns.importedNamespaces()):
                    self.__namespaceGraph.addEdge(ns, rns)
                    if not rns in done_check:
                        need_check.add(rns)
                if not ns.hasSchemaComponents():
                    _log.warning('Referenced %s has no schema components', ns.uri())
                done_check.add(ns)
            assert done_check == self.__namespaceGraph.nodes()

        return self.__namespaceGraph
    __namespaceGraph = None

    def namespaceOrder (self, reset=False):
        return self.namespaceGraph(reset).sccOrder()

    def siblingsFromGraph (self, reset=False):
        siblings = set()
        ns_graph = self.namespaceGraph(reset)
        for ns in self.__rootNamespaces:
            ns_siblings = ns_graph.sccMap().get(ns)
            if ns_siblings is not None:
                siblings.update(ns_siblings)
            else:
                siblings.add(ns)
        return siblings

    def siblingNamespaces (self):
        if self.__siblingNamespaces is None:
            self.__siblingNamespaces = self.siblingsFromGraph()
        return self.__siblingNamespaces

    def setSiblingNamespaces (self, sibling_namespaces):
        self.__siblingNamespaces = sibling_namespaces

    __siblingNamespaces = None

    def dependentNamespaces (self, reset=False):
        return self.namespaceGraph(reset).nodes()

    def componentGraph (self, reset=False):
        if reset or (self.__componentGraph is None):
            self.__componentGraph = pyxb.utils.utility.Graph()
            all_components = set()
            for ns in self.siblingNamespaces():
                [ all_components.add(_c) for _c in ns.components() if _c.hasBinding() ]

            need_visit = all_components.copy()
            while 0 < len(need_visit):
                c = need_visit.pop()
                self.__componentGraph.addNode(c)
                for cd in c.bindingRequires(include_lax=True):
                    if cd in all_components:
                        self.__componentGraph.addEdge(c, cd)
        return self.__componentGraph
    __componentGraph = None

    def componentOrder (self, reset=False):
        return self.componentGraph(reset).sccOrder()

    def __init__ (self, **kw):
        namespace_set = set(kw.get('namespace_set', []))
        namespace = kw.get('namespace')
        if namespace is not None:
            namespace_set.add(namespace)
        if 0 == len(namespace_set):
            raise pyxb.LogicError('NamespaceDependencies requires at least one root namespace')
        self.__rootNamespaces = namespace_set


## Local Variables:
## fill-column:78
## End:



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
# Declare xml namespace
import pyxb.binding.generate
import pyxb.utils.domutils
from xml.dom import Node

import os.path

xsd='''<?xml version="1.0" encoding="UTF-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:xml="http://www.w3.org/XML/1998/namespace">
  <xs:complexType name="structure">
    <xs:attributeGroup ref="xml:specialAttrs"/>
  </xs:complexType>
  <xs:element name="instance" type="structure"/>
</xs:schema>'''

code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
#open('code.py', 'w').write(code)

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

class TestTrac0023 (unittest.TestCase):
    def testBasic (self):
        self.assertEqual(4, len(structure._AttributeMap))
        self.assertEqual(pyxb.binding.xml_.STD_ANON_lang, structure._AttributeMap[pyxb.namespace.XML.createExpandedName('lang')].dataType())
        self.assertEqual(pyxb.binding.xml_.STD_ANON_space, structure._AttributeMap[pyxb.namespace.XML.createExpandedName('space')].dataType())
        self.assertEqual(pyxb.binding.datatypes.anyURI, structure._AttributeMap[pyxb.namespace.XML.createExpandedName('base')].dataType())
        self.assertEqual(pyxb.binding.datatypes.ID, structure._AttributeMap[pyxb.namespace.XML.createExpandedName('id')].dataType())

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.utils.domutils
from pyxb.utils import six
from xml.dom import Node

import os.path
xst = '''<?xml version="1.0" encoding="UTF-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
  <xs:simpleType name="int32">
      <xs:restriction base="xs:hexBinary">
          <xs:length value="4"/>
      </xs:restriction>
  </xs:simpleType>
  <xs:attribute name="Color" use="optional" type="int32"/>
  <xs:element name="elt">
    <xs:complexType>
      <xs:simpleContent>
        <xs:restriction base="int32">
          <xs:attribute ref="Color"/>
        </xs:restriction>
      </xs:simpleContent>
    </xs:complexType>
  </xs:element>
</xs:schema>
'''

code = pyxb.binding.generate.GeneratePython(schema_text=xst)
#print code

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

class TestTrac0216 (unittest.TestCase):

    def testBasic (self):
        xmlt = six.u('<elt>30313233</elt>')
        xmld = xmlt.encode('utf-8')
        instance = CreateFromDocument(xmlt)
        self.assertEqual(b'0123', instance.value())
        xmlt = six.u('<elt Color="33323130">30313233</elt>')
        xmld = xmlt.encode('utf-8')
        instance = CreateFromDocument(xmlt)
        self.assertEqual(b'0123', instance.value())
        self.assertEqual(b'3210', instance.Color)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.utils.domutils
from pyxb.utils import six
from xml.dom import Node
import pyxb.namespace
import xml.dom.minidom as minidom

import os.path
xst = '''<?xml version="1.0" encoding="utf-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
  <xs:element name="elt" type="xs:string"/>
</xs:schema>
'''


code = pyxb.binding.generate.GeneratePython(schema_text=xst)
#open('code.py', 'w').write(code)

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

import copy

class TestIssue19 (unittest.TestCase):
    def testCopy (self):
        x = elt('shallow')
        # This one doesn't generate a warning
        x2 = copy.copy(x)
        self.assertEqual(x, x2)
        self.assertNotEqual(id(x), id(x2))
        self.assertEqual('shallow', x2)

    def testDeepCopy (self):
        x = elt('deep')
        # This one does generate a warning
        x2 = copy.deepcopy(x)
        self.assertEqual(x, x2)
        self.assertNotEqual(id(x), id(x2))
        self.assertEqual('deep', x2)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.utils.domutils
from pyxb.utils import six
from xml.dom import Node

import os.path
xst = '''<?xml version="1.0"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
  <xs:element name="anything">
    <xs:complexType mixed="true">
      <xs:sequence>
        <xs:any minOccurs="0"/>
      </xs:sequence>
    </xs:complexType>
  </xs:element>
  <xs:element name="text" type="xs:string"/>
</xs:schema>
'''

code = pyxb.binding.generate.GeneratePython(schema_text=xst)
#print code

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

class TestTrac0219 (unittest.TestCase):
    def testElement (self):
        xmlt = six.u('<anything><text>eltcontent</text></anything>')
        instance = CreateFromDocument(xmlt)
        self.assertEqual(1, len(instance.wildcardElements()))
        oc = instance.orderedContent()
        self.assertEqual(1, len(oc))
        self.assertTrue(isinstance(oc[0], pyxb.binding.basis.ElementContent))
        i = anything()
        i.append(text('eltcontent'))
        i.validateBinding()
        xmlt = six.u('<anything><text>eltcontent</text></anything>')
        xmld = xmlt.encode('utf-8')
        self.assertEqual(xmld, i.toxml('utf-8', root_only=True))

    def testMixedOnly (self):
        xmlt = six.u('<anything>mixed</anything>')
        instance = CreateFromDocument(xmlt)
        self.assertEqual(0, len(instance.wildcardElements()))
        oc = instance.orderedContent()
        self.assertEqual(1, len(oc))
        self.assertTrue(isinstance(oc[0], pyxb.binding.basis.NonElementContent))
        i = anything()
        i.append('mixed')
        i.validateBinding()
        xmlt = six.u('<anything>mixed</anything>')
        xmld = xmlt.encode('utf-8')
        self.assertEqual(xmld, i.toxml('utf-8', root_only=True))

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import unittest
from pyxb.utils.unicode import *

class TestCodePointSet (unittest.TestCase):
    def testConstructor (self):
        c = CodePointSet()
        self.assertEqual(c.asTuples(), [])
        c = CodePointSet(10, 15)
        self.assertEqual(c.asTuples(), [ (10, 10), (15, 15) ])
        c = CodePointSet([10, 15])
        self.assertEqual(c.asTuples(), [ (10, 10), (15, 15) ])

    def testCopyConstructor (self):
        c = CodePointSet()
        c.add(10)
        c.add(15)
        self.assertEqual(c.asTuples(), [ (10, 10), (15, 15) ])
        c2 = CodePointSet(c)
        self.assertEqual(c2.asTuples(), [ (10, 10), (15, 15) ])
        c.add(20)
        self.assertEqual(c.asTuples(), [ (10, 10), (15, 15), (20, 20) ])
        self.assertEqual(c2.asTuples(), [ (10, 10), (15, 15) ])

    def testNegate (self):
        c = CodePointSet().negate()
        self.assertEqual(c.asTuples(), [ (0, CodePointSet.MaxCodePoint) ])
        c2 = c.negate()
        self.assertEqual(c2.asTuples(), [])

    def testAddSingle (self):
        c = CodePointSet()
        c.add(15)
        self.assertEqual(c.asTuples(), [ (15, 15) ])
        c.add(15)
        self.assertEqual(c.asTuples(), [ (15, 15) ])

        c.add(0)
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15) ])
        n = c.negate()
        self.assertEqual(n.asTuples(), [ (1, 14), (16, CodePointSet.MaxCodePoint) ])

        n.add(0)
        self.assertEqual(n.asTuples(), [ (0, 14), (16, CodePointSet.MaxCodePoint) ])
        n.add(15)
        self.assertEqual(n.asTuples(), [ (0, CodePointSet.MaxCodePoint) ])

        c = CodePointSet()
        c.add(':')
        self.assertEqual(c.asTuples(), [ (58, 58) ])

    def testRemoveRange (self):
        base = CodePointSet(0, 15, (20, 30), (40, 60))
        self.assertEqual(base.asTuples(), [ (0, 0), (15, 15), (20, 30), (40, 60) ])
        # 0 1 15 16 20 31 40 61
        c = CodePointSet(base).subtract( (22, 25) )
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 21), (26, 30), (40, 60) ])
        c = CodePointSet(base).subtract( (22, 35) )
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 21), (40, 60) ])
        c = CodePointSet(base).subtract( (35, 55) )
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 30), (56, 60) ])
        c = CodePointSet(base).subtract( (35, 38) )
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 30), (40, 60) ])

    def testAddRange (self):
        base = CodePointSet(0, 15)
        self.assertEqual(base.asTuples(), [ (0, 0), (15, 15) ])
        base.add((20, 30))
        self.assertEqual(base.asTuples(), [ (0, 0), (15, 15), (20, 30) ])
        base.add((40, 60))
        self.assertEqual(base.asTuples(), [ (0, 0), (15, 15), (20, 30), (40, 60) ])
        # 0 1 15 16 20 31 40 61
        # Bridge missing range
        c = CodePointSet(base).add((1, 15))
        self.assertEqual(c.asTuples(), [ (0, 15), (20, 30), (40, 60) ])

        # Insert in middle of missing range
        c = CodePointSet(base).add((35, 38))
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 30), (35, 38), (40, 60) ])
        # Join following range
        c = CodePointSet(base).add((35, 39))
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 30), (35, 60) ])
        c = CodePointSet(base).add((35, 40))
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 30), (35, 60) ])

        # 0 1 15 16 20 31 40 61
        # Insert into middle of existing range
        c = CodePointSet(base).add((22, 25))
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 30), (40, 60) ])
        # Extend existing range
        c = CodePointSet(base).add((22, 35))
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 35), (40, 60) ])
        c = CodePointSet(base).add((22, 38))
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 38), (40, 60) ])
        # Span missing range
        c = CodePointSet(base).add((22, 39))
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 60) ])
        c = CodePointSet(base).add((22, 40))
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 60) ])
        c = CodePointSet(base).add((22, 41))
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 60) ])

        # 0 1 15 16 20 31 40 61
        c = CodePointSet(base).add((15, 18))
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 18), (20, 30), (40, 60) ])
        c = CodePointSet(base).add((35, 65))
        self.assertEqual(c.asTuples(), [ (0, 0), (15, 15), (20, 30), (35, 65) ])
        c = CodePointSet(base).add((12, 16))
        self.assertEqual(c.asTuples(), [ (0, 0), (12, 16), (20, 30), (40, 60) ])

    def testAsPattern (self):
        c = CodePointSet()
        self.assertEqual('[]', c.asPattern())
        c.add(':')
        self.assertEqual('[\u003A]', c.asPattern())

        n = c.negate()
        if SupportsWideUnicode:
            self.assertEqual('[\\x00-\u0039\u003B-\U0010FFFF]', n.asPattern())
        else:
            self.assertEqual('[\\x00-\u0039\u003B-\uFFFF]', n.asPattern())

        c = CodePointSet()
        c.add(']')
        self.assertEqual('[\]]', c.asPattern())
        c = CodePointSet()
        c.add('-')
        c.add('+')
        self.assertEqual('[+\-]', c.asPattern())


    def testAsSingleCharacter (self):
        c = CodePointSet()
        self.assertTrue(c.asSingleCharacter() is None)
        c.add('A')
        self.assertEqual('A', c.asSingleCharacter())
        c.add('B')
        self.assertTrue(c.asSingleCharacter() is None)
        self.assertEqual('G', CodePointSet('G').asSingleCharacter())
        self.assertEqual('\u0041', CodePointSet(65).asSingleCharacter())
        self.assertEqual('\uFFFF', CodePointSet(0xFFFF).asSingleCharacter())

class TestXML1p0e2 (unittest.TestCase):
    def testChar (self):
        if SupportsWideUnicode:
            self.assertTrue(( 1+CodePointSet.MaxShortCodePoint, CodePointSet.MaxCodePoint ) in XML1p0e2.Char.asTuples())

if '__main__' == __name__:
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.utils.domutils
from pyxb.utils import six
import pyxb.utils.utility
from pyxb.utils.utility import MakeIdentifier

import os.path
xsd='''<?xml version="1.0" encoding="UTF-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
   <xs:complexType name="simple_type">
      <xs:simpleContent>
         <xs:extension base="xs:string">
           <xs:attribute name="is_clean" type="xs:boolean"/>
         </xs:extension>
      </xs:simpleContent>
   </xs:complexType>
   <xs:element name="simple_element" type="simple_type"/>
</xs:schema>'''

code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
rv = compile(code, 'test', 'exec')
eval(rv)

import unittest
import re

_CamelCase_re = re.compile(r'_\w')

def MakeCamelCase (identifier):
    return _CamelCase_re.sub(lambda _m: _m.group(0)[1].upper(), six.text_type(identifier))

class TestTrac0141 (unittest.TestCase):
    def tearDown (cls):
        pyxb.utils.utility._SetXMLIdentifierToPython(None)

    def testDefaultMakeIdentifier (self):
        self.assertEqual(MakeIdentifier('is_string'), 'is_string')

    def testReplacedMakeIdentifier (self):
        pyxb.utils.utility._SetXMLIdentifierToPython(MakeCamelCase)
        self.assertEqual(MakeIdentifier('is_string'), 'isString')

    def testReplacedBuild (self):
        pyxb.utils.utility._SetXMLIdentifierToPython(MakeCamelCase)
        code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
        rv = compile(code, 'test', 'exec')
        xglobals = globals().copy()
        xlocals = locals().copy()
        self.assertFalse('simpleElement' in xlocals)
        eval(rv, xglobals, xlocals)
        self.assertTrue('simpleElement' in xlocals)
        instance = xlocals['simpleElement'](isClean=True)
        self.assertTrue(instance.isClean)

    def testNormalBuild (self):
        instance = simple_element(is_clean=True)
        self.assertTrue(instance.is_clean)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.utils.domutils
from xml.dom import Node

import os.path
xsd='''<?xml version="1.0" encoding="UTF-8"?>
 <xsd:schema xmlns:xsd="http://www.w3.org/2001/XMLSchema">

  <xsd:complexType name="tAny">
    <xsd:all minOccurs="0">
      <xsd:element type="xsd:int" name="a" minOccurs="1"/>
      <xsd:element type="xsd:int" name="b" minOccurs="1"/>
      <xsd:element type="xsd:int" name="c" minOccurs="1"/>
      <xsd:element type="xsd:int" name="d" minOccurs="1"/>
    </xsd:all>
  </xsd:complexType>
  <xsd:element name="eAny" type="tAny"/>
</xsd:schema>'''

code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
#open('code.py', 'w').write(code)

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

class TestTrac0181 (unittest.TestCase):
    def getInstance (self):
        instance = eAny()
        vc = instance._validationConfig.copy()
        instance._setValidationConfig(vc)
        instance.a = 1
        instance.d = 2
        instance.c = 3
        instance.b = 4
        return instance

    def testNEVER (self):
        i = self.getInstance()
        vc = i._validationConfig
        vc._setContentInfluencesGeneration(vc.NEVER)
        # Uses declaration order for sub-automata (alphabetic)
        xmls = '<eAny><a>1</a><b>4</b><c>3</c><d>2</d></eAny>'
        xmld = xmls.encode('utf-8')
        self.assertEqual(i.toxml('utf-8', root_only=True), xmld)

    def testALWAYS (self):
        i = self.getInstance()
        vc = i._validationConfig
        vc._setContentInfluencesGeneration(vc.ALWAYS)
        # Uses assignment order for sub-automata (numeric)
        xmls = '<eAny><a>1</a><d>2</d><c>3</c><b>4</b></eAny>'
        xmld = xmls.encode('utf-8')
        self.assertEqual(i.toxml('utf-8', root_only=True), xmld)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.utils.domutils
from pyxb.utils import six
from xml.dom import Node

import os.path
xsd='''<?xml version="1.0" encoding="UTF-8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
  <xs:group name="gText">
    <xs:choice>
      <xs:element ref="text"/>
      <xs:element ref="bold"/>
      <xs:element ref="ital"/>
    </xs:choice>
  </xs:group>
  <xs:complexType name="tText" mixed="true">
    <xs:group ref="gText" minOccurs="0" maxOccurs="unbounded"/>
  </xs:complexType>
  <xs:complexType name="tBold" mixed="true">
    <xs:group ref="gText" minOccurs="0" maxOccurs="unbounded"/>
  </xs:complexType>
  <xs:complexType name="tItal" mixed="true">
    <xs:group ref="gText" minOccurs="0" maxOccurs="unbounded"/>
  </xs:complexType>
  <xs:element name="text" type="tText"/>
  <xs:element name="bold" type="tBold"/>
  <xs:element name="ital" type="tItal"/>
  <xs:complexType name="tOrdered">
    <xs:sequence>
      <xs:element ref="bold" minOccurs="0" maxOccurs="unbounded"/>
      <xs:element ref="ital" minOccurs="0" maxOccurs="unbounded"/>
      <xs:element ref="text" minOccurs="0" maxOccurs="unbounded"/>
    </xs:sequence>
  </xs:complexType>
  <xs:element name="ordered" type="tOrdered"/>
</xs:schema>'''

code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
#open('code.py', 'w').write(code)

rv = compile(code, 'test', 'exec')
eval(rv)

# Assign a shared validation configuration to these types
gvc = pyxb.GlobalValidationConfig
vc = gvc.copy()
for cls in [ tText, tBold, tItal, tOrdered ]:
    cls._SetValidationConfig(vc)

from pyxb.exceptions_ import *

import unittest
import sys

class TestTrac0153 (unittest.TestCase):
    def tearDown (self):
        vc._setContentInfluencesGeneration(gvc.contentInfluencesGeneration)
        vc._setOrphanElementInContent(gvc.orphanElementInContent)
        vc._setInvalidElementInContent(gvc.invalidElementInContent)

    ExpectedText = '''<text>Intro with <bold>bold text</bold> and <ital>italicized text with <bold>bold</bold> inside</ital> ending with a little <bold>more bold</bold> text.</text>'''
    ExpectedData = ExpectedText.encode('utf-8')

    def makeText (self):
        return text('Intro with ',
                    bold('bold text'),
                    ' and ',
                    ital('italicized text with ', bold('bold'), ' inside'),
                    ' ending with a little ',
                    bold('more bold'),
                    ' text.')

    ExpectedMultit = '''<text>t1<bold>b2</bold>t3<bold>b4</bold><ital>i5</ital><bold>b6</bold></text>'''
    ExpectedMultid = ExpectedMultit.encode('utf-8')
    def makeMulti (self):
        return text('t1', bold('b2'), 't3', bold('b4'), ital('i5'), bold('b6'))

    def testDefaults (self):
        self.assertEqual(vc.contentInfluencesGeneration, vc.MIXED_ONLY)
        self.assertEqual(vc.orphanElementInContent, vc.IGNORE_ONCE)
        self.assertEqual(vc.invalidElementInContent, vc.IGNORE_ONCE)

    def testMakeText (self):
        i = self.makeText()
        self.assertEqual(2, len(i.bold))
        self.assertEqual(1, len(i.ital))
        self.assertEqual(1, len(i.ital[0].bold))
        self.assertEqual(self.ExpectedData, i.toxml('utf-8', root_only=True))
        i2 = CreateFromDocument(self.ExpectedText)
        self.assertEqual(self.ExpectedData, i2.toxml('utf-8', root_only=True))

    def testNeverCIT (self):
        i = self.makeText()
        vc._setContentInfluencesGeneration(vc.NEVER)
        # All non-element content is lost, and element content is
        # emitted in declaration order.
        xmlt = six.u('<text><bold/><bold/><ital><bold/></ital></text>')
        xmld = xmlt.encode('utf-8')
        self.assertEqual(i.toxml('utf-8', root_only=True), xmld)

    def testOrphan (self):
        i = self.makeText()
        # Drop the second bold
        dropped = i.bold.pop()
        self.assertEqual(vc.orphanElementInContent, vc.IGNORE_ONCE)
        xmlt = six.u('<text>Intro with <bold>bold text</bold> and <ital>italicized text with <bold>bold</bold> inside</ital> ending with a little  text.</text>')
        xmld = xmlt.encode('utf-8')
        self.assertEqual(i.toxml('utf-8', root_only=True), xmld)
        vc._setOrphanElementInContent(vc.GIVE_UP)
        self.assertEqual(vc.orphanElementInContent, vc.GIVE_UP)
        self.assertEqual(gvc.orphanElementInContent, gvc.IGNORE_ONCE)
        xmlt = six.u('<text>Intro with <bold>bold text</bold> and <ital>italicized text with <bold>bold</bold> inside</ital> ending with a little  text.</text>')
        xmld = xmlt.encode('utf-8')
        self.assertEqual(i.toxml('utf-8', root_only=True), xmld)
        vc._setOrphanElementInContent(vc.RAISE_EXCEPTION)
        self.assertEqual(vc.orphanElementInContent, vc.RAISE_EXCEPTION)
        self.assertEqual(gvc.orphanElementInContent, gvc.IGNORE_ONCE)
        if sys.version_info[:2] < (2, 7):
            self.assertRaises(pyxb.OrphanElementContentError, i.toxml, 'utf-8', root_only=True)
            return
        with self.assertRaises(pyxb.OrphanElementContentError) as cm:
            xmld = i.toxml('utf-8', root_only=True)
        e = cm.exception
        self.assertEqual(e.instance, i)
        self.assertEqual(e.preferred.value, dropped)

    def testOrphan2 (self):
        i = self.makeMulti()
        xmld = i.toxml('utf-8', root_only=True)
        self.assertEqual(self.ExpectedMultid, xmld)
        self.assertEqual(3, len(i.bold))
        dropped = i.bold.pop(0)
        self.assertEqual(vc.orphanElementInContent, vc.IGNORE_ONCE)
        xmlt = six.u('<text>t1t3<bold>b4</bold><ital>i5</ital><bold>b6</bold></text>')
        xmld = xmlt.encode('utf-8')
        self.assertEqual(i.toxml('utf-8', root_only=True), xmld)
        vc._setOrphanElementInContent(vc.GIVE_UP)
        # Elements in declaration order, non-element content at end
        xmlt = six.u('<text><bold>b4</bold><bold>b6</bold><ital>i5</ital>t1t3</text>')
        xmld = xmlt.encode('utf-8')
        self.assertEqual(i.toxml('utf-8', root_only=True), xmld)

    ExpectedOrderedt = '''<ordered><bold>b1</bold><bold>b2</bold><ital>i1</ital><text>t1</text></ordered>'''
    ExpectedOrderedd = ExpectedOrderedt.encode('utf-8')

    def makeOrdered (self):
        return ordered(bold('b1'), bold('b2'), ital('i1'), text('t1'))

    def testOrdered (self):
        i = self.makeOrdered()
        xmld = i.toxml('utf-8', root_only=True)
        self.assertEqual(self.ExpectedOrderedd, xmld)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import unittest
from pyxb.utils.saxutils import *
from xml.dom import Node
import xml.dom
import pyxb.namespace

class TestState (SAXElementState):
    StateSequence = []

    def __init__ (self, *args, **kw):
        self.StateSequence.append(self)
        super(TestState, self).__init__(*args, **kw)

BogusNamespace = pyxb.namespace.NamespaceInstance('urn:test-saxutils-bogus')
books_ns = pyxb.namespace.NamespaceInstance('urn:loc.gov:books')
isbn_ns = pyxb.namespace.NamespaceInstance('urn:ISBN:0-395-36341-6')
xhtml_ns = pyxb.namespace.NamespaceInstance('http://www.w3.org/1999/xhtml')

class TestInScopeNames (unittest.TestCase):
    def show (self, node):
        xmlns_map = pyxb.namespace.NamespaceContext.GetNodeContext(node).inScopeNamespaces()
        #print '%s: %s' % (node.nodeName, ' ; '.join([ '%s=%s' % (_k, _v.uri()) for (_k, _v) in six.iteritems(xmlns_map)]))
        return xmlns_map

    def tearDown (self):
        TestState.StateSequence[:] = []

    def stripUndeclaredNamespaces (self, xmlns_map):
        xmlns_map = xmlns_map.copy()
        self.assertEqual(pyxb.namespace.XML.uri(), xmlns_map.pop('xml').uri())
        self.assertEqual(pyxb.namespace.XMLNamespaces.uri(), xmlns_map.pop('xmlns').uri())
        return xmlns_map

    def test_6_2_2 (self):
        xmld = '''<?xml version="1.0"?>
<!-- initially, the default namespace is "books" -->
<book xmlns='urn:loc.gov:books'
      xmlns:isbn='urn:ISBN:0-395-36341-6'>
    <title>Cheaper by the Dozen</title>
    <isbn:number>1568491379</isbn:number>
    <notes>
      <p xmlns='http://www.w3.org/1999/xhtml'>
          This is a <i>funny</i> book!
      </p>
      <p>another graf without namespace change</p>
    </notes>
</book>'''.encode('utf-8')
        saxer = make_parser(element_state_constructor=TestState, location_base='test_6_2_2', fallback_namespace=BogusNamespace)
        handler = saxer.getContentHandler()
        saxer.parse(io.BytesIO(xmld))

        # First is root context; second is the book element
        book = TestState.StateSequence[1]
        en = book.expandedName()
        self.assertTrue(en is not None)
        self.assertEqual(en.namespace(), books_ns)
        self.assertEqual(en.localName(), 'book')
        self.assertEqual(book.namespaceContext().defaultNamespace(), books_ns)
        self.assertEqual(book.namespaceContext().inScopeNamespaces().get('isbn'), isbn_ns)

        title = TestState.StateSequence[2]
        xmlns_map = self.stripUndeclaredNamespaces(title.namespaceContext().inScopeNamespaces())
        self.assertEqual(2, len(xmlns_map))
        self.assertEqual('urn:loc.gov:books', xmlns_map[None].uri())
        self.assertEqual('urn:ISBN:0-395-36341-6', xmlns_map['isbn'].uri())

        p = TestState.StateSequence[5]
        xmlns_map = self.stripUndeclaredNamespaces(p.namespaceContext().inScopeNamespaces())
        self.assertEqual(p.expandedName().localName(), 'p')
        self.assertEqual(2, len(xmlns_map))
        self.assertEqual('http://www.w3.org/1999/xhtml', xmlns_map[None].uri())
        self.assertEqual('urn:ISBN:0-395-36341-6', xmlns_map['isbn'].uri())

        x = TestState.StateSequence[7]
        xmlns_map = self.stripUndeclaredNamespaces(x.namespaceContext().inScopeNamespaces())
        self.assertEqual(x.expandedName().localName(), 'p')
        self.assertEqual(2, len(xmlns_map))
        self.assertEqual('urn:loc.gov:books', xmlns_map[None].uri())
        self.assertEqual('urn:ISBN:0-395-36341-6', xmlns_map['isbn'].uri())

    def test_6_2_3 (self):
        xmld = '''<?xml version='1.0'?>
<Beers>
  <table xmlns='http://www.w3.org/1999/xhtml'>
   <th><td>Name</td><td>Origin</td><td>Description</td></th>
   <tr>
     <td><brandName xmlns="">Huntsman</brandName></td>
     <td><origin xmlns="">Bath, UK</origin></td>
     <td>
       <details xmlns=""><class>Bitter</class><hop>Fuggles</hop>
         <pro>Wonderful hop, light alcohol, good summer beer</pro>
         <con>Fragile; excessive variance pub to pub</con>
         </details>
        </td>
      </tr>
    </table>
  </Beers>'''.encode('utf-8')

        saxer = make_parser(element_state_constructor=TestState, location_base='test_6_2_3', fallback_namespace=BogusNamespace)
        handler = saxer.getContentHandler()
        saxer.parse(io.BytesIO(xmld))

        Beers = TestState.StateSequence[1]
        xmlns_map = self.stripUndeclaredNamespaces(Beers.namespaceContext().inScopeNamespaces())
        self.assertEqual(0, len(xmlns_map))
        table = TestState.StateSequence[2]
        self.assertEqual(xhtml_ns.createExpandedName('table'), table.expandedName())
        xmlns_map = self.stripUndeclaredNamespaces(table.namespaceContext().inScopeNamespaces())
        self.assertEqual(1, len(xmlns_map))
        self.assertEqual('http://www.w3.org/1999/xhtml', xmlns_map[None].uri())
        brandName = TestState.StateSequence[9]
        xmlns_map = self.stripUndeclaredNamespaces(brandName.namespaceContext().inScopeNamespaces())
        self.assertTrue(brandName.expandedName().namespace() is None)
        self.assertEqual('brandName', brandName.expandedName().localName())
        self.assertEqual(0, len(xmlns_map))

if '__main__' == __name__:
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import unittest

import root
import branch1
import branch2

class TestTrac0183 (unittest.TestCase):

    XMLS = '''<ns:msg xmlns:ns="urn:root" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <e3>re3</e3>
  <e1 xmlns="urn:branch1">b1e1</e1>
  <e2 xmlns="urn:branch2">b2e2</e2>
</ns:msg>'''

    def testBasic (self):
        instance = root.CreateFromDocument(self.XMLS)
        self.assertEqual('b1e1', instance.e1)
        self.assertTrue(instance.e1._element(), branch1.e1)
        self.assertEqual('b2e2', instance.e2)
        self.assertTrue(instance.e2._element(), branch2.e2)
        self.assertEqual('re3', instance.e3)
        self.assertTrue(instance.e3._element(), root.msg.typeDefinition()._UseForTag('e3'))

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
if __name__ == '__main__':
    logging.basicConfig()
_log = logging.getLogger(__name__)
import pyxb.binding.generate
import pyxb.utils.domutils
from xml.dom import Node

import os.path
xsd='''<?xml version="1.0" encoding="UTF-8"?>
<xs:schema  xmlns:xs="http://www.w3.org/2001/XMLSchema">
  <xs:element name="elt1" type="xs:string"/>
  <xs:element name="elt2" type="xs:string"/>
  <xs:complexType name="mytype">
     <xs:sequence>
       <xs:element ref="elt1"/>
       <xs:element ref="elt2"/>
     </xs:sequence>
  </xs:complexType>
  <xs:complexType name="mytypeWC">
     <xs:sequence>
       <xs:element ref="elt1"/>
       <xs:element ref="elt2"/>
       <xs:any namespace="##other" processContents="lax"/>
     </xs:sequence>
  </xs:complexType>
</xs:schema>
'''

code = pyxb.binding.generate.GeneratePython(schema_text=xsd)
#print code

rv = compile(code, 'test', 'exec')
eval(rv)

from pyxb.exceptions_ import *

import unittest

class TestTrac0029 (unittest.TestCase):
    """Presence of a wildcard in a sequence model group causes other
    elements in that group to not be generated."""
    def test (self):
        self.assertTrue(mytype._UseForTag('elt1') is not None)
        self.assertTrue(mytype._UseForTag('elt2') is not None)

        self.assertTrue(mytypeWC._UseForTag('elt1') is not None)
        self.assertTrue(mytypeWC._UseForTag('elt2') is not None)


if __name__ == '__main__':
    unittest.main()
